from typing import Any, Dict, List, Optional

import requests
from requests.auth import HTTPBasicAuth

from autumn8.cli.cli_environment import CliEnvironment
from autumn8.common.config.settings import CloudServiceProvider
from autumn8.common.types import Sla
from autumn8.lib.api_creds import retrieve_api_creds
from autumn8.lib.http import require_ok_response, url_with_params
from autumn8.types.deployment import DeploymentView

DEFAULT_API_TIMEOUT = 60


def get_running_deployments(
    organization_id: int,
    environment: CliEnvironment,
    model_id: Optional[int] = None,
    service_provider: Optional[CloudServiceProvider] = None,
) -> List[DeploymentView]:
    autodl_host = environment.value.app_host

    params: Dict[str, Any] = {"organization_id": organization_id}
    if model_id is not None:
        params["model_id"] = model_id
    if service_provider is not None:
        params["service_provider"] = service_provider

    deployments_api_route = f"{autodl_host}/api/cloud/deployments"
    # TODO: wrap the requests library in a custom class to handle common logic like auth and headers for json
    response = requests.get(
        url_with_params(deployments_api_route, params),
        headers={"Content-Type": "application/json"},
        auth=HTTPBasicAuth(*retrieve_api_creds()),
        timeout=DEFAULT_API_TIMEOUT,
    )

    require_ok_response(response)
    return [
        DeploymentView.parse_obj(item)
        for item in response.json()["deployments"]
    ]


def deploy_by_best_sla(
    organization_id: int,
    environment: CliEnvironment,
    model_id: int,
    best_sla: Sla,
    schedule_on: Optional[str],
    deployment_id: Optional[str],
    service_provider: CloudServiceProvider,
):
    autodl_host = environment.value.app_host

    params = {
        "organization_id": organization_id,
        "model_id": model_id,
        "best_sla": best_sla.value,
        "schedule_on": schedule_on,
        "deployment_id": deployment_id,
        "cloud_service_provider": service_provider.value,
    }

    print("sending", params)

    deployments_api_route = f"{autodl_host}/api/cloud/deployments/by_sla"
    response = requests.post(
        url_with_params(deployments_api_route, params),
        headers={"Content-Type": "application/json"},
        auth=HTTPBasicAuth(*retrieve_api_creds()),
        timeout=DEFAULT_API_TIMEOUT,
    )

    require_ok_response(response)
    return response.json()


def deploy(
    organization_id: int,
    environment: CliEnvironment,
    model_id: int,
    machine_type: str,
    schedule_on: Optional[str],
    deployment_id: Optional[str],
    service_provider: CloudServiceProvider,
):
    autodl_host = environment.value.app_host

    params = {
        "organization_id": organization_id,
        "model_id": model_id,
        "machine_type": machine_type,
        "schedule_on": schedule_on,
        "deployment_id": deployment_id,
        "cloud_service_provider": service_provider.value,
    }

    deployments_api_route = f"{autodl_host}/api/cloud/deployments"
    response = requests.post(
        url_with_params(deployments_api_route, params),
        headers={"Content-Type": "application/json"},
        auth=HTTPBasicAuth(*retrieve_api_creds()),
        timeout=DEFAULT_API_TIMEOUT,
    )

    require_ok_response(response)
    return response.json()


def terminate_deployment(
    organization_id: int,
    environment: CliEnvironment,
    deployment_id: str,
    service_provider: CloudServiceProvider,
):
    autodl_host = environment.value.app_host

    params = {
        "organization_id": organization_id,
        "service_provider": service_provider.value,
    }

    deployments_api_route = (
        f"{autodl_host}/api/cloud/deployments/{deployment_id}"
    )
    response = requests.delete(
        url_with_params(deployments_api_route, params),
        headers={"Content-Type": "application/json"},
        auth=HTTPBasicAuth(*retrieve_api_creds()),
        timeout=DEFAULT_API_TIMEOUT,
    )

    require_ok_response(response)
    return response.json()



import datetime
import json
import os
from typing import Optional

import jsonschema
import jsonschema.exceptions
import questionary
from jsonschema import validate
from prompt_toolkit.document import Document


class IsoDatetimeValidator(questionary.Validator):
    @classmethod
    def validate_string(cls, maybe_iso_datetime_string: Optional[str]):
        if maybe_iso_datetime_string is None or maybe_iso_datetime_string == "":
            raise questionary.ValidationError(message="Input is empty")

        try:
            date = datetime.datetime.fromisoformat(
                maybe_iso_datetime_string
            ).astimezone()
            if date < datetime.datetime.now().astimezone():
                raise questionary.ValidationError(
                    message="Date/time entered has already passed"
                )

        except ValueError as exc:
            raise questionary.ValidationError(
                message="Input is not a valid ISO format time string"
            ) from exc

    @classmethod
    def is_valid_iso_datetime_string(
        cls, maybe_iso_datetime_string: str
    ) -> bool:
        return cls.get_maybe_error_message(maybe_iso_datetime_string) is None

    @classmethod
    def get_maybe_error_message(
        cls, maybe_iso_datetime_string: str
    ) -> Optional[str]:
        try:
            IsoDatetimeValidator.validate_string(maybe_iso_datetime_string)
            return None
        except questionary.ValidationError as exc:
            return exc.message

    def validate(self, document: Document) -> None:
        IsoDatetimeValidator.validate_string("\n".join(document.lines))


INPUT_DIMS_JSONSCHEMA = {
    "type": "array",
    "minItems": 1,
    "items": {
        "type": "array",
        "minItems": 1,
        "items": {"type": "number"},
    },
}


def validate_input_dims_json(json_string: str):
    if json_string == "":
        return True

    try:
        json_data = json.loads(json_string)
        validate(instance=json_data, schema=INPUT_DIMS_JSONSCHEMA)
    except (
        jsonschema.exceptions.ValidationError,
        json.decoder.JSONDecodeError,
    ):
        return False
    return True


def validate_input_file(path: str):
    if not os.path.exists(path):
        return False

    if not os.path.isfile(path):
        return False

    try:
        with open(path, "r") as f:
            json_data = json.load(f)

    except json.decoder.JSONDecodeError:
        return False
    return True



import numpy as np
import torch
from sentence_transformers import util
from sentence_transformers.SentenceTransformer import SentenceTransformer
from transformers import AutoModel, AutoTokenizer

import autumn8

query = "How many people live in London?"
docs = [
    "Around 9 Million people live in London",
    "London is known for its financial district",
]

# Load the model
model = SentenceTransformer("sentence-transformers/multi-qa-mpnet-base-dot-v1")
tokenizer = AutoTokenizer.from_pretrained(
    "sentence-transformers/multi-qa-mpnet-base-dot-v1"
)


def preprocess(texts):
    query, docs = texts
    encoded_query = tokenizer(
        query, padding=True, truncation=True, return_tensors="pt"
    )
    encoded_docs = tokenizer(
        docs, padding=True, truncation=True, return_tensors="pt"
    )
    return (encoded_query, encoded_docs)


def model_func(input):
    query, docs = input
    doc_emb = model(docs)["sentence_embedding"]
    query_emb = model(query)["sentence_embedding"]
    return (query_emb, doc_emb)


def postprocess(model_output):
    query_emb, doc_emb = model_output
    # Compute dot score between query and all document embeddings
    scores = util.dot_score(query_emb, doc_emb)[0].cpu().tolist()

    # Combine docs & scores
    doc_score_pairs = list(zip(docs, scores))

    # Sort by decreasing score
    doc_score_pairs = sorted(doc_score_pairs, key=lambda x: x[1], reverse=True)

    # Output passages & scores
    for doc, score in doc_score_pairs:
        print(score, doc)

    return doc_score_pairs


autumn8.lib.attach_model(
    model_func,
    (query, docs),
    preprocess=preprocess,
    postprocess=postprocess,
    interns=[],
    externs=[
        "torch",
        "torchvision",
        "sentence_transformers",
        "sentence_transformers.SentenceTransformer",
        "sentence_transformers.util",
        "jinja",
        "huggingface_hub",
        "yaml",
        "transformers.models.mpnet.modeling_mpnet",
        "tensorflow",
        "transformers.activations",
        "transformers.models.mpnet.configuration_mpnet",
        "transformers.models.mpnet.tokenization_mpnet_fast",
        "tokenizers",
        "tokenizers.models",
        "requests",
        "numpy",
        "transformers",
        "tqdm.autonotebook",
        "sys",
        "packaging.version",
        "requests.exceptions",
        "filelock",
        "tqdm",
        "tqdm.auto",
        "_io",
        "requests.auth",
        "tensorboard.summary._tf.summary",
    ],
)



"""A client service for CloudInfo, reqires CloudInfo to be running and available under CLOUDINFO_API_PATH"""
# https://github.com/banzaicloud/cloudinfo

import os
from dataclasses import dataclass
from http import HTTPStatus
from typing import Dict, Literal, Optional, Union

import libenv
import requests
from fastapi import HTTPException

from autumn8.common.config.settings import CloudServiceProvider

CPU_ARCHITECTURE = Literal["x86_64", "arm64"]

SERVICE_PROVIDER_NAME_TRANSLATION_DICT = {
    "alibaba": "alibaba",
    "amazon": "amazon",
    "azure": "azure",
    "digitalocean": "digitalocean",
    "google cloud platform": "google",
    "oracle": "oracle",
}


@dataclass
class InstanceDescription:
    label: str
    family: str
    instance_description_link: str
    hw: str
    cores: int
    hyperthreading: bool
    usd_per_hr: Optional[float]
    predictor_target: str
    predictor_num_threads: float
    service_provider: CloudServiceProvider
    ram: Optional[float]
    gpuram: Optional[float]
    is_supported_by_a8f: bool = False
    cpu_arch: Optional[CPU_ARCHITECTURE] = None

    # you can determine if the instance is available in govcloud
    # using https://calculator.aws/#/addService/ec2-enhancement
    # and checking either of the GovCloud regions
    aws_govcloud: Union[bool, str] = False


def map_autodl_service_provider_to_cloudinfo_provider(
    autodl_service_provider: str,
) -> str:
    # based off of https://github.com/banzaicloud/cloudinfo/tree/master/internal/cloudinfo/providers
    if (
        autodl_service_provider.lower()
        not in SERVICE_PROVIDER_NAME_TRANSLATION_DICT
    ):
        raise ValueError(
            f"Unsupported service provider {autodl_service_provider}"
        )

    return SERVICE_PROVIDER_NAME_TRANSLATION_DICT[
        autodl_service_provider.lower()
    ]


class CloudInfoService:
    _cloud_products_info = None

    def __init__(
        self,
        provider: str,
        region: str,
        fetch_data_from_cloud_info: bool = True,
    ) -> None:
        self.cloudinfo_service_enabled = fetch_data_from_cloud_info
        if (
            provider.lower()
            not in SERVICE_PROVIDER_NAME_TRANSLATION_DICT.keys()
        ):
            raise RuntimeError("Invalid provider")
        self.provider = provider
        self.region = region

    @property
    def cloud_products_info(self):
        """Get cloud products info and cache it for further reuse"""

        if not self.cloudinfo_service_enabled:
            return None

        if self._cloud_products_info is None:
            service_type = "compute"
            cloudinfo_api_path = libenv.CLOUDINFO_API_PATH
            provider = map_autodl_service_provider_to_cloudinfo_provider(
                self.provider
            )

            url = f"{cloudinfo_api_path}/providers/{provider}/services/{service_type}/regions/{self.region}/products"

            response = requests.get(
                url,
                headers={"Content-Type": "application/json"},
            )
            if not response.ok:
                if "status not yet cached" in response.text:
                    raise HTTPException(
                        status_code=HTTPStatus.SERVICE_UNAVAILABLE,
                        detail=f"Try again later, cloudinfo service is still caching machine statuses",
                    )
                raise ValueError(
                    f"Could not get cloud products info from {url}: {response.text}"
                )

            product_descriptions_list = response.json()["products"]

            self._cloud_products_info = {
                product["type"]: product
                for product in product_descriptions_list
            }

        return self._cloud_products_info

    def get_instance_info(self, instance_label):
        if instance_label not in self.cloud_products_info:
            raise (
                ValueError(
                    f"Could not find cloud product of type {instance_label}"
                )
            )
        return self.cloud_products_info[instance_label]

    def get_instance_pricing(self, instance_label):
        if not self.cloudinfo_service_enabled:
            return None

        instance_info = self.get_instance_info(
            instance_label=instance_label,
        )
        return instance_info["onDemandPrice"]

    def build_instance_description(
        self,
        *,
        label: str,
        family: str,
        instance_description_link: str,
        hw: str,
        cores: int,
        hyperthreading: bool,
        hardcoded_usd_per_hr: Optional[float] = None,
        predictor_target: str,
        predictor_num_threads: float,
        ram: float,
        gpuram: Optional[float] = None,
        is_supported_by_a8f: bool = False,
        aws_govcloud: Union[bool, str] = False,
        cpu_arch: Optional[CPU_ARCHITECTURE] = None,
    ) -> InstanceDescription:
        usd_per_hr = (
            hardcoded_usd_per_hr
            if hardcoded_usd_per_hr
            else self.get_instance_pricing(label)
        )

        # TODO rename variables to align with their meaning better
        return InstanceDescription(
            label=label,
            family=family,
            instance_description_link=instance_description_link,
            hw=hw,
            cores=cores,
            hyperthreading=hyperthreading,
            usd_per_hr=usd_per_hr,
            predictor_target=predictor_target,
            predictor_num_threads=predictor_num_threads,
            service_provider=self.provider,
            ram=ram,
            gpuram=gpuram,
            is_supported_by_a8f=is_supported_by_a8f,
            aws_govcloud=aws_govcloud,
            cpu_arch=cpu_arch,
        )

    def build_keyed_instance_description(
        self,
        *,
        label: str,
        family: str,
        instance_description_link: str,
        hw: str,
        cores: int,
        hyperthreading: bool,
        hardcoded_usd_per_hr: Optional[float] = None,
        predictor_target: str,
        predictor_num_threads: int,
        ram: float,
        gpuram: Optional[float] = None,
        is_supported_by_a8f: bool = False,
        aws_govcloud: Union[bool, str] = False,
        cpu_arch: Optional[CPU_ARCHITECTURE] = None,
    ) -> Dict[str, Dict[int, InstanceDescription]]:
        return {
            predictor_target: {
                predictor_num_threads: self.build_instance_description(
                    label=label,
                    family=family,
                    instance_description_link=instance_description_link,
                    hw=hw,
                    cores=cores,
                    hyperthreading=hyperthreading,
                    hardcoded_usd_per_hr=hardcoded_usd_per_hr,
                    predictor_target=predictor_target,
                    predictor_num_threads=predictor_num_threads,
                    ram=ram,
                    gpuram=gpuram,
                    is_supported_by_a8f=is_supported_by_a8f,
                    aws_govcloud=aws_govcloud,
                    cpu_arch=cpu_arch,
                )
            }
        }



import asyncio
import datetime
import json
import re
from functools import wraps
from typing import Optional, Tuple, TypedDict, Union

import click
import questionary
from questionary import Choice

import autumn8.lib.api.cloud as cloud_api
from autumn8.cli.validation import IsoDatetimeValidator
from autumn8.lib import logging
from autumn8.lib.api.user import fetch_user_data
from autumn8.types.deployment import DeploymentView

logger = logging.getLogger(__name__)


def normalize_args(name: str):
    """
    Use this with a click.group to allow both underscores and dashes
    in the CLI flags

    For example, this will make the CLI allow both --model_id and --model-id

    Patch stolen from https://github.com/pallets/click/issues/1123#issuecomment-589989721
    """
    return name.replace("_", "-")


def get_user_organizations(user_data):
    return [mem["organization"] for mem in user_data["memberships"]]


def pick_organization_id(environment):
    user_data = fetch_user_data(environment)
    user_organizations = get_user_organizations(user_data)

    organization_id = questionary.select(
        "Choose organization",
        choices=[
            Choice(title=f"{org['name']} ({org['id']})", value=org["id"])
            for org in user_organizations
        ],
        use_shortcuts=True,
    ).unsafe_ask()
    return organization_id


def verify_organization_id_access(environment, organization_id):
    user_data = fetch_user_data(environment)
    user_organization_ids = [
        org["id"] for org in get_user_organizations(user_data)
    ]
    if organization_id not in user_organization_ids:
        raise Exception(
            f"The user {user_data['email']} does not belong to the organization of id={organization_id}"
        )


def render_deployment(deployment: DeploymentView):
    return (
        f"{deployment.status} model {deployment.model_id} "
        + f"on {deployment.service_provider} {deployment.instance_type} "
        + f"(id={deployment.deployment_id})"
    )


def get_deployment(organization_id, environment, model_id=None):
    deployments = cloud_api.get_running_deployments(
        organization_id,
        environment,
        model_id=model_id,
    )

    if len(deployments) == 1:
        return deployments[0].public_dns
    else:
        choices = [
            Choice(
                title=render_deployment(deployment),
                value=deployment.public_dns,
            )
            for deployment in deployments
        ]
        question = questionary.select(
            "Choose deployment",
            choices=choices,
            use_shortcuts=(len(choices) < 36),
            use_arrow_keys=True,
        )
        return question.unsafe_ask()


def announce_model_upload_response(model_upload_response):
    return announce_json_response({"model_details": model_upload_response})


def announce_json_response(model_upload_response):
    logger.info("")  # newline
    logger.info("Done!")
    click.echo(json.dumps(model_upload_response, indent=4))


def validate_and_ask_about_schedule(
    should_schedule: bool, schedule_on: Optional[str]
) -> Optional[str]:
    if schedule_on is not None:
        try:
            IsoDatetimeValidator.validate_string(schedule_on)
            return schedule_on
        except questionary.ValidationError as exc:
            click.echo(
                f"ERROR: '{schedule_on}' is not a valid date specification"
            )
            click.echo(exc.message + "\n")

    if should_schedule or schedule_on is not None:
        prefill_value = (
            datetime.datetime.now().astimezone().replace(microsecond=0)
            + datetime.timedelta(minutes=5)
        ).isoformat()

        return questionary.text(
            "Pick time to schedule the model to deploy on\n"
            + f"(ie. in 5 minutes it will be '{prefill_value}')\n",
            default=schedule_on or prefill_value,
            validate=IsoDatetimeValidator,
        ).unsafe_ask()

    return None


async def validate_and_ask_about_docker_image_name(docker_image_name: str):
    # https://regex101.com/r/tPWcN4/1
    regex = re.compile(
        r"^(?P<repository>[\w.\-_]+(?:(?::\d+|)(?=/[a-z0-9._-]+/[a-z0-9._-]+))|)(?:/|)(?P<image>[a-z0-9.\-_]+(?:/[a-z0-9.\-_]+|):?[\w.\-_]{1,127})$",
        re.I,
    )
    if not regex.fullmatch(docker_image_name):
        docker_image_name = await questionary.text(
            "Pick docker image to host",
            default=docker_image_name or "",
            validate=lambda input: bool(regex.fullmatch(input)),
        ).unsafe_ask_async()
    final_match = regex.fullmatch(docker_image_name)
    if not final_match:
        raise RuntimeError("Impossible case (?)")
    match_groups: tuple[str] = final_match.groups()

    repository, image = match_groups
    return repository, image


# https://github.com/pallets/click/issues/85
def coro_click_command(func):
    """
    Make a Click command run in an asyncio compatible context

    Breaks synchronous questionary, use async instead
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        return asyncio.run(func(*args, **kwargs))

    return wrapper



import dataclasses
import enum
import json
from dataclasses import dataclass
from typing import Optional, Union

from pydantic import BaseModel


class DeploymentStatus(str, enum.Enum):
    STARTING = "STARTING"
    READY = "READY"
    INVALID = "INVALID"
    CRASHED = "CRASHED"
    SLEEPING = "SLEEPING"
    TERMINATING = "TERMINATING"
    TERMINATED = "TERMINATED"


class ServerType(str, enum.Enum):
    MULTIMODEL_SERVER = "MULTIMODEL_SERVER"
    CUSTOM_PYTORCH_INFERER = "PYTORCH"  # for CLI imported models
    TORCHSERVE = "TORCHSERVE"
    TENSORFLOW_SERVING = "TENSORFLOW"


class DeploymentView(BaseModel):
    initialization_progress: int = 100
    """Percents of progress, in range from 0 to 100"""
    initialization_status_details: str = (
        "Deployment set up and accepting requests"
    )

    deployment_id: str
    """Unique identifier of the deployment"""
    model_id: int
    """Unique identifier of the model that's hosted by the deployment"""
    instance_type: str

    status: DeploymentStatus = DeploymentStatus.STARTING
    created_at: Union[int, None] = None
    public_dns: Union[str, None] = None
    server_type: Union[ServerType, None] = None
    service_provider: str = "autumn8"


@dataclass
class A8FManifest:
    version: str
    """The version of the manifest"""

    uri: str
    """S3 Object key, or HTTP url, under which the model file is available"""

    memory: int
    model_id: int
    """AutoDL ID of the model"""

    model_name: str
    deployment_id: Optional[
        str
    ] = None  # idk why, but for some time we passed this only for deployments with external s3 bucket models
    model_bucket: Optional[str] = None
    """An external public S3 bucket containing the model file"""

    @classmethod
    def load(cls, json_string: str) -> "A8FManifest":
        dict_content = json.loads(json_string)
        return cls(**dict_content)

    def dict(self):
        return dataclasses.asdict(self)



import posixpath

import requests
from requests.auth import HTTPBasicAuth

from autumn8.cli.cli_environment import CliEnvironment
from autumn8.lib import logging
from autumn8.lib.api_creds import retrieve_api_creds
from autumn8.lib.http import require_ok_response

logger = logging.getLogger(__name__)


def user_api_url(environment: CliEnvironment, subpath: str) -> str:
    autodl_host = environment.value.app_host
    if subpath:
        return f"{autodl_host}/api/user/{subpath}"
    return f"{autodl_host}/api/user"


def fetch_user_data(environment: CliEnvironment):
    user_id, api_key = retrieve_api_creds()

    response = requests.get(
        user_api_url(environment, ""),
        headers={"Content-Type": "application/json"},
        auth=HTTPBasicAuth(user_id, api_key),
    )

    require_ok_response(response)
    return response.json()["user"]


def fetch_org_api_keys(environment: CliEnvironment, organization_id: int):
    user_id, api_key = retrieve_api_creds()
    autodl_host = environment.value.app_host

    response = requests.get(
        user_api_url(environment, f"inference_api_keys"),
        params={"organization_id": organization_id},
        headers={"Content-Type": "application/json"},
        auth=HTTPBasicAuth(user_id, api_key),
    )
    require_ok_response(response)
    return response.json()["api_keys"]



from http import HTTPStatus
from typing import Callable, Dict, Optional

from fastapi import HTTPException

from autumn8.common.config.cloud_info import (
    CloudInfoService,
    InstanceDescription,
)
from autumn8.common.config.settings import CloudServiceProvider
from autumn8.common.types import DEFAULT_CLOUD_ZONES_CONFIG, ZoneConfig
from autumn8.lib import logging

logger = logging.getLogger(__name__)


def get_default_cloud_info_services_per_provider():
    return {
        provider: CloudInfoService(provider, zone)
        for provider, zone in DEFAULT_CLOUD_ZONES_CONFIG.items()
    }


def find_instance_config(
    predictor_target: str,
    fetch_data_from_cloud_info: bool = True,
    num_threads: Optional[int] = None,
) -> InstanceDescription:
    instances = get_supported_cloud_instances(
        fetch_data_from_cloud_info=fetch_data_from_cloud_info
    )

    if predictor_target not in instances:
        raise HTTPException(
            HTTPStatus.NOT_FOUND,
            f'Instance config not found for instance_type="{predictor_target}"',
        )

    instances_by_num_threads = instances[predictor_target]

    if len(instances_by_num_threads.items()) == 1:
        the_only_key = [*instances_by_num_threads.keys()][0]
        num_threads = the_only_key

    if not num_threads:
        # TODO: after replacing predictor_target with label, remove this check
        raise ValueError(
            f"You have to provide num_threads to be able to query Oracle instances"
        )

    return instances_by_num_threads[num_threads]


# TODO: use some smarter caching mechanism here
ALL_CLOUD_INFO_SERVICES = {}

PLACEHOLDER_ZONE_FOR_DISCONNECTED_CLOUD_INFO_SERVICES = "disconnected"


def get_cloud_info_service(
    provider: CloudServiceProvider,
    zone_config: ZoneConfig,
    fetch_data_from_cloud_info=True,
):
    if provider not in ALL_CLOUD_INFO_SERVICES:
        ALL_CLOUD_INFO_SERVICES[provider] = {}

    provider_services = ALL_CLOUD_INFO_SERVICES[provider]
    zone = (
        zone_config[provider]
        if fetch_data_from_cloud_info
        else PLACEHOLDER_ZONE_FOR_DISCONNECTED_CLOUD_INFO_SERVICES
    )

    if zone not in provider_services:
        provider_services[zone] = CloudInfoService(
            provider,
            zone,
            fetch_data_from_cloud_info=fetch_data_from_cloud_info,
        )

    return provider_services[zone]


INSTANCE_DESCRIPTIONS_PER_ZONE = {}


def get_instances_for_cloud_provider(
    provider: CloudServiceProvider,
    zone_config: ZoneConfig,
    instances_getter: Callable,
    fetch_data_from_cloud_info=True,
):
    cloud_info_service = get_cloud_info_service(
        provider,
        zone_config,
        fetch_data_from_cloud_info=fetch_data_from_cloud_info,
    )
    zone = (
        zone_config[provider]
        if fetch_data_from_cloud_info
        else PLACEHOLDER_ZONE_FOR_DISCONNECTED_CLOUD_INFO_SERVICES
    )

    if provider not in INSTANCE_DESCRIPTIONS_PER_ZONE:
        INSTANCE_DESCRIPTIONS_PER_ZONE[provider] = {}

    provider_instances = INSTANCE_DESCRIPTIONS_PER_ZONE[provider]

    if zone not in provider_instances:
        provider_instances[zone] = instances_getter(cloud_info_service)

    return provider_instances[zone]


INSTANCE_RECOMMENDATIONS_LABELS = [
    "t3.medium",
    "c5.large",
    "g5.2xlarge",
    "e2-standard-2",
    "c2-standard-4",
    "VM.Standard.A1.Flex 2vCPU",
    "Azure.StandardB2s",
]


def get_recommendations(fetch_data_from_cloud_info=True):
    all_supported_instances = get_supported_cloud_instances(
        fetch_data_from_cloud_info=fetch_data_from_cloud_info
    )
    instance_descriptions = [
        description
        for instances_by_thread_count in all_supported_instances.values()
        for description in instances_by_thread_count.values()
    ]
    return [
        instance
        for instance in instance_descriptions
        if instance.label in INSTANCE_RECOMMENDATIONS_LABELS
    ]


# TODO gracefully handle cases where a service is not available in a specific zone
# TODO users should be able to choose cheaper regions
# ASSUMPTION: pair (predictor_target, predictor_num_threads) is unique across
#   all instance_types, across all cloud service_providers
# TODO: patch Oracle metrics to use label instead of the duplicate
#   predictor_target, drop num_threads from unique constraint
def get_supported_cloud_instances(
    zone_config=DEFAULT_CLOUD_ZONES_CONFIG,
    fetch_data_from_cloud_info=True,
) -> Dict[str, Dict[float, InstanceDescription]]:

    return {
        # TODO: replace predictor_target with 'label' instead, eliminate
        #   second dict layer indexing by num_threads - run migration
        #   for all instances, rename benchmarks
        **get_instances_for_cloud_provider(
            CloudServiceProvider.AMAZON,
            zone_config,
            get_supported_amazon_cloud_instances,
            fetch_data_from_cloud_info=fetch_data_from_cloud_info,
        ),
        **get_instances_for_cloud_provider(
            CloudServiceProvider.ORACLE,
            zone_config,
            get_supported_oracle_cloud_instances,
            fetch_data_from_cloud_info=fetch_data_from_cloud_info,
        ),
        **get_instances_for_cloud_provider(
            CloudServiceProvider.GOOGLE,
            zone_config,
            get_supported_google_cloud_instances,
            fetch_data_from_cloud_info=fetch_data_from_cloud_info,
        ),
        **get_instances_for_cloud_provider(
            CloudServiceProvider.AZURE,
            zone_config,
            get_supported_azure_cloud_instances,
            fetch_data_from_cloud_info=fetch_data_from_cloud_info,
        ),
    }


def get_supported_amazon_cloud_instances(
    amazon_cloud_info_service: CloudInfoService,
):
    return {
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="lambda_10240MB",
            family="lambda",
            instance_description_link="https://aws.amazon.com/lambda/",
            hw="",
            cores=5,
            hyperthreading=True,
            hardcoded_usd_per_hr=0.0000001667
            * 1000
            * 3600,  # TODO cloudinfo doesn't support lambdas pricing, handle that across zones somehow
            predictor_target="lambda_10240MB",
            predictor_num_threads=5,
            ram=10,
            is_supported_by_a8f=True,
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c5.large",
            family="c5",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c5/",
            hw="Intel Xeon",
            cores=2,
            hyperthreading=True,
            predictor_target="c5.large",
            predictor_num_threads=1,
            ram=4,
            is_supported_by_a8f=True,
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c5.xlarge",
            family="c5",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c5/",
            hw="Intel Xeon",
            cores=4,
            hyperthreading=True,
            predictor_target="c5.xlarge",
            predictor_num_threads=2,
            ram=8,
            is_supported_by_a8f=True,
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c5.2xlarge",
            family="c5",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c5/",
            hw="Intel Xeon",
            cores=8,
            hyperthreading=True,
            predictor_target="c5.2xlarge",
            predictor_num_threads=4,
            ram=16,
            is_supported_by_a8f=True,
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c5.4xlarge",
            family="c5",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c5/",
            hw="Intel Xeon",
            cores=16,
            hyperthreading=True,
            predictor_target="c5.4xlarge",
            predictor_num_threads=8,
            ram=32,
            is_supported_by_a8f=True,
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c5.9xlarge",
            family="c5",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c5/",
            hw="Intel Xeon",
            cores=36,
            hyperthreading=True,
            predictor_target="c5.9xlarge",
            predictor_num_threads=18,
            ram=72,
            is_supported_by_a8f=True,
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c5.12xlarge",
            family="c5",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c5/",
            hw="Intel Xeon",
            cores=48,
            hyperthreading=True,
            predictor_target="c5.12xlarge",
            predictor_num_threads=24,
            ram=96,
            is_supported_by_a8f=True,
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c5.18xlarge",
            family="c5",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c5/",
            hw="Intel Xeon",
            cores=72,
            hyperthreading=True,
            predictor_target="c5.18xlarge",
            predictor_num_threads=36,
            ram=144,
            is_supported_by_a8f=True,
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c5.24xlarge",
            family="c5",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c5/",
            hw="Intel Xeon",
            cores=96,
            hyperthreading=True,
            predictor_target="c5.24xlarge",
            predictor_num_threads=48,
            ram=192,
            is_supported_by_a8f=True,
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c5a.large",
            family="c5a",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c5/",
            hw="AMD EPYC",
            cores=2,
            hyperthreading=True,
            predictor_target="c5a.large",
            predictor_num_threads=1,
            ram=4,
            is_supported_by_a8f=True,
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c5a.xlarge",
            family="c5a",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c5/",
            hw="AMD EPYC",
            cores=4,
            hyperthreading=True,
            predictor_target="c5a.xlarge",
            predictor_num_threads=2,
            ram=8,
            is_supported_by_a8f=True,
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c5a.2xlarge",
            family="c5a",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c5/",
            hw="AMD EPYC",
            cores=8,
            hyperthreading=True,
            predictor_target="c5a.2xlarge",
            predictor_num_threads=4,
            ram=16,
            is_supported_by_a8f=True,
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c5a.4xlarge",
            family="c5a",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c5/",
            hw="AMD EPYC",
            cores=16,
            hyperthreading=True,
            predictor_target="c5a.4xlarge",
            predictor_num_threads=8,
            ram=32,
            is_supported_by_a8f=True,
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c5a.8xlarge",
            family="c5a",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c5/",
            hw="AMD EPYC",
            cores=32,
            hyperthreading=True,
            predictor_target="c5a.8xlarge",
            predictor_num_threads=16,
            ram=64,
            is_supported_by_a8f=True,
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c5a.12xlarge",
            family="c5a",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c5/",
            hw="AMD EPYC",
            cores=48,
            hyperthreading=True,
            predictor_target="c5a.12xlarge",
            predictor_num_threads=24,
            ram=96,
            is_supported_by_a8f=True,
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c5a.16xlarge",
            family="c5a",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c5/",
            hw="AMD EPYC",
            cores=64,
            hyperthreading=True,
            predictor_target="c5a.16xlarge",
            predictor_num_threads=32,
            ram=128,
            is_supported_by_a8f=True,
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c5a.24xlarge",
            family="c5a",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c5/",
            hw="AMD EPYC",
            cores=96,
            hyperthreading=True,
            predictor_target="c5a.24xlarge",
            predictor_num_threads=48,
            ram=192,
            is_supported_by_a8f=True,
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c5n.large",
            family="c5n",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c5/",
            hw="Intel Xeon",
            cores=2,
            hyperthreading=True,
            predictor_target="c5n.large",
            predictor_num_threads=2,
            ram=5.25,
            is_supported_by_a8f=True,
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c5n.xlarge",
            family="c5n",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c5/",
            hw="Intel Xeon",
            cores=4,
            hyperthreading=True,
            predictor_target="c5n.xlarge",
            predictor_num_threads=4,
            ram=10.5,
            is_supported_by_a8f=True,
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c5n.2xlarge",
            family="c5n",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c5/",
            hw="Intel Xeon",
            cores=8,
            hyperthreading=True,
            predictor_target="c5n.2xlarge",
            predictor_num_threads=8,
            ram=21,
            is_supported_by_a8f=True,
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c5n.4xlarge",
            family="c5n",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c5/",
            hw="Intel Xeon",
            cores=16,
            hyperthreading=True,
            predictor_target="c5n.4xlarge",
            predictor_num_threads=16,
            ram=42,
            is_supported_by_a8f=True,
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c5n.9xlarge",
            family="c5n",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c5/",
            hw="Intel Xeon",
            cores=36,
            hyperthreading=True,
            predictor_target="c5n.9xlarge",
            predictor_num_threads=36,
            ram=96,
            is_supported_by_a8f=True,
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c5n.18xlarge",
            family="c5n",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c5/",
            hw="Intel Xeon",
            cores=72,
            hyperthreading=True,
            predictor_target="c5n.18xlarge",
            predictor_num_threads=72,
            ram=192,
            is_supported_by_a8f=True,
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="t2.small",
            family="t2",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/t2/",
            hw="AMD EPYC",
            cores=1,
            hyperthreading=True,
            predictor_target="t2.small",
            predictor_num_threads=1,
            ram=2,
            is_supported_by_a8f=True,
            aws_govcloud="US West only",
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="t2.medium",
            family="t2",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/t2/",
            hw="AMD EPYC",
            cores=2,
            hyperthreading=True,
            predictor_target="t2.medium",
            predictor_num_threads=2,
            ram=4,
            is_supported_by_a8f=True,
            aws_govcloud="US West only",
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="t2.large",
            family="t2",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/t2/",
            hw="AMD EPYC",
            cores=2,
            hyperthreading=True,
            predictor_target="t2.large",
            predictor_num_threads=2,
            ram=8,
            is_supported_by_a8f=True,
            aws_govcloud="US West only",
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="t2.xlarge",
            family="t2",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/t2/",
            hw="AMD EPYC",
            cores=4,
            hyperthreading=True,
            predictor_target="t2.xlarge",
            predictor_num_threads=4,
            ram=16,
            is_supported_by_a8f=True,
            aws_govcloud="US West only",
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="t2.2xlarge",
            family="t2",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/t2/",
            hw="AMD EPYC",
            cores=8,
            hyperthreading=True,
            predictor_target="t2.2xlarge",
            predictor_num_threads=8,
            ram=32,
            is_supported_by_a8f=True,
            aws_govcloud="US West only",
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="t3.small",
            family="t3",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/t3/",
            hw="Intel Xeon Platinum",
            cores=2,
            hyperthreading=True,
            predictor_target="t3.small",
            predictor_num_threads=2,
            ram=2,
            is_supported_by_a8f=True,
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="t3.medium",
            family="t3",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/t3/",
            hw="Intel Xeon Platinum",
            cores=2,
            hyperthreading=True,
            predictor_target="t3.medium",
            predictor_num_threads=2,
            ram=4,
            is_supported_by_a8f=True,
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="t3.large",
            family="t3",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/t3/",
            hw="Intel Xeon Platinum",
            cores=2,
            hyperthreading=True,
            predictor_target="t3.large",
            predictor_num_threads=2,
            ram=8,
            is_supported_by_a8f=True,
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="t3.xlarge",
            family="t3",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/t3/",
            hw="Intel Xeon Platinum",
            cores=4,
            hyperthreading=True,
            predictor_target="t3.xlarge",
            predictor_num_threads=4,
            ram=16,
            is_supported_by_a8f=True,
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="t3.2xlarge",
            family="t3",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/t3/",
            hw="Intel Xeon Platinum",
            cores=8,
            hyperthreading=True,
            predictor_target="t3.2xlarge",
            predictor_num_threads=8,
            ram=32,
            is_supported_by_a8f=True,
            aws_govcloud=True,
        ),
        # t4g micro and nano missing
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="t4g.small",
            family="t4",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/t4/",
            hw="Amazon Graviton2",
            cores=2,
            hyperthreading=True,
            predictor_target="t4g.small",
            predictor_num_threads=4,
            ram=2,
            cpu_arch="arm64",
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="t4g.medium",
            family="t4",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/t4/",
            hw="Amazon Graviton2",
            cores=2,
            hyperthreading=True,
            predictor_target="t4g.medium",
            predictor_num_threads=4,
            ram=4,
            cpu_arch="arm64",
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="t4g.large",
            family="t4",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/t4/",
            hw="Amazon Graviton2",
            cores=2,
            hyperthreading=True,
            predictor_target="t4g.large",
            predictor_num_threads=4,
            ram=8,
            cpu_arch="arm64",
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="t4g.xlarge",
            family="t4",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/t4/",
            hw="Amazon Graviton2",
            cores=4,
            hyperthreading=True,
            predictor_target="t4g.xlarge",
            predictor_num_threads=8,
            ram=16,
            cpu_arch="arm64",
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="t4g.2xlarge",
            family="t4",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/t4/",
            hw="Amazon Graviton2",
            cores=8,
            hyperthreading=True,
            predictor_target="t4g.2xlarge",
            predictor_num_threads=16,
            ram=32,
            cpu_arch="arm64",
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c6g.medium",
            family="c6g",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c6g/",
            hw="Amazon Graviton2",
            cores=1,
            hyperthreading=False,
            predictor_target="c6g.medium",
            predictor_num_threads=1,
            ram=2,
            cpu_arch="arm64",
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c6g.large",
            family="c6g",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c6g/",
            hw="Amazon Graviton2",
            cores=2,
            hyperthreading=False,
            predictor_target="c6g.large",
            predictor_num_threads=2,
            ram=4,
            cpu_arch="arm64",
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c6g.xlarge",
            family="c6g",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c6g/",
            hw="Amazon Graviton2",
            cores=4,
            hyperthreading=False,
            predictor_target="c6g.xlarge",
            predictor_num_threads=4,
            ram=8,
            cpu_arch="arm64",
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c6g.2xlarge",
            family="c6g",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c6g/",
            hw="Amazon Graviton2",
            cores=8,
            hyperthreading=False,
            predictor_target="c6g.2xlarge",
            predictor_num_threads=8,
            ram=16,
            cpu_arch="arm64",
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c6g.4xlarge",
            family="c6g",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c6g/",
            hw="Amazon Graviton2",
            cores=16,
            hyperthreading=False,
            predictor_target="c6g.4xlarge",
            predictor_num_threads=16,
            ram=32,
            cpu_arch="arm64",
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c6g.8xlarge",
            family="c6g",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c6g/",
            hw="Amazon Graviton2",
            cores=32,
            hyperthreading=False,
            predictor_target="c6g.8xlarge",
            predictor_num_threads=32,
            ram=64,
            cpu_arch="arm64",
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c6g.12xlarge",
            family="c6g",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c6g/",
            hw="Amazon Graviton2",
            cores=48,
            hyperthreading=False,
            predictor_target="c6g.12xlarge",
            predictor_num_threads=48,
            ram=96,
            cpu_arch="arm64",
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c6g.16xlarge",
            family="c6g",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c6g/",
            hw="Amazon Graviton2",
            cores=64,
            hyperthreading=False,
            predictor_target="c6g.16xlarge",
            predictor_num_threads=64,
            ram=128,
            cpu_arch="arm64",
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c6gn.medium",
            family="c6gn",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c6g/",
            hw="Amazon Graviton2",
            cores=1,
            hyperthreading=False,
            predictor_target="c6gn.medium",
            predictor_num_threads=1,
            ram=2,
            cpu_arch="arm64",
            aws_govcloud=False,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c6gn.large",
            family="c6gn",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c6g/",
            hw="Amazon Graviton2",
            cores=2,
            hyperthreading=False,
            predictor_target="c6gn.large",
            predictor_num_threads=2,
            ram=4,
            cpu_arch="arm64",
            aws_govcloud=False,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c6gn.xlarge",
            family="c6gn",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c6g/",
            hw="Amazon Graviton2",
            cores=4,
            hyperthreading=False,
            predictor_target="c6gn.xlarge",
            predictor_num_threads=4,
            ram=8,
            cpu_arch="arm64",
            aws_govcloud=False,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c6gn.2xlarge",
            family="c6gn",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c6g/",
            hw="Amazon Graviton2",
            cores=8,
            hyperthreading=False,
            predictor_target="c6gn.2xlarge",
            predictor_num_threads=8,
            ram=16,
            cpu_arch="arm64",
            aws_govcloud=False,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c6gn.4xlarge",
            family="c6gn",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c6g/",
            hw="Amazon Graviton2",
            cores=16,
            hyperthreading=False,
            predictor_target="c6gn.4xlarge",
            predictor_num_threads=16,
            ram=32,
            cpu_arch="arm64",
            aws_govcloud=False,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c6gn.8xlarge",
            family="c6gn",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c6g/",
            hw="Amazon Graviton2",
            cores=32,
            hyperthreading=False,
            predictor_target="c6gn.8xlarge",
            predictor_num_threads=32,
            ram=64,
            cpu_arch="arm64",
            aws_govcloud=False,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c6gn.12xlarge",
            family="c6gn",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c6g/",
            hw="Amazon Graviton2",
            cores=48,
            hyperthreading=False,
            predictor_target="c6gn.12xlarge",
            predictor_num_threads=48,
            ram=96,
            cpu_arch="arm64",
            aws_govcloud=False,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c6gn.16xlarge",
            family="c6gn",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/c6g/",
            hw="Amazon Graviton2",
            cores=64,
            hyperthreading=False,
            predictor_target="c6gn.16xlarge",
            predictor_num_threads=64,
            ram=128,
            cpu_arch="arm64",
            aws_govcloud=False,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c4.large",
            family="c4",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/",
            hw="Intel Xeon",
            cores=2,
            hyperthreading=True,
            predictor_target="c4.large",
            predictor_num_threads=2,
            ram=3.75,
            aws_govcloud="US West only",
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c4.xlarge",
            family="c4",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/",
            hw="Intel Xeon",
            cores=4,
            hyperthreading=True,
            predictor_target="c4.xlarge",
            predictor_num_threads=4,
            ram=7.5,
            aws_govcloud="US West only",
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c4.2xlarge",
            family="c4",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/",
            hw="Intel Xeon",
            cores=8,
            hyperthreading=True,
            predictor_target="c4.2xlarge",
            predictor_num_threads=8,
            ram=15,
            aws_govcloud="US West only",
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c4.4xlarge",
            family="c4",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/",
            hw="Intel Xeon",
            cores=16,
            hyperthreading=True,
            predictor_target="c4.4xlarge",
            predictor_num_threads=16,
            ram=30,
            aws_govcloud="US West only",
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="c4.8xlarge",
            family="c4",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/",
            hw="Intel Xeon",
            cores=36,
            hyperthreading=True,
            predictor_target="c4.8xlarge",
            predictor_num_threads=36,
            ram=60,
            aws_govcloud="US West only",
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="g4ad.2xlarge",
            family="g4ad",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/g4/",
            hw="NVIDIA T4 GPUs",
            cores=8,
            hyperthreading=False,
            hardcoded_usd_per_hr=0.752,
            predictor_target="g4ad.2xlarge",
            predictor_num_threads=8,
            ram=32,
            gpuram=16,
            # TODO: gradually enable instances on staging while adding more a8f support
            is_supported_by_a8f=False,
            aws_govcloud=False,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="p3.2xlarge",
            family="p3",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/p3/",
            hw="NVIDIA V100 GPUs",
            cores=8,
            hyperthreading=False,
            hardcoded_usd_per_hr=3.06,
            predictor_target="p3.2xlarge",
            predictor_num_threads=8,
            ram=61,
            gpuram=16,
            is_supported_by_a8f=True,
            aws_govcloud="US West only",
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="p2.xlarge",
            family="p2",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/p2/",
            hw="NVIDIA K80 GPUs",
            cores=4,
            hyperthreading=False,
            predictor_target="p2.xlarge",
            predictor_num_threads=4,
            ram=61,
            gpuram=24,
            is_supported_by_a8f=False,
            aws_govcloud="US West only",
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="p3.8xlarge",
            family="p3",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/p3/",
            hw="NVIDIA V100 GPUs",
            cores=32,
            hyperthreading=False,
            predictor_target="p3.8xlarge",
            predictor_num_threads=16,
            ram=244,
            gpuram=64,
            is_supported_by_a8f=True,
            aws_govcloud="US West only",
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="g3s.xlarge",
            family="g3s",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/g3/",
            hw="NVIDIA Tesla M60 GPU",
            cores=2,
            hyperthreading=False,
            predictor_target="g3s.xlarge",
            predictor_num_threads=4,
            ram=30.5,
            gpuram=16,
            is_supported_by_a8f=True,
            aws_govcloud="US West only",
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="g4dn.xlarge",
            family="g4dn",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/g4/",
            hw="NVIDIA T4 GPU",
            cores=4,
            hyperthreading=False,
            predictor_target="g4dn.xlarge",
            predictor_num_threads=4,
            ram=16,
            gpuram=16,
            is_supported_by_a8f=True,
            aws_govcloud=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="g5.xlarge",
            family="g5",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/g5/",
            hw="NVIDIA A10G",
            cores=4,
            hyperthreading=False,
            predictor_target="g5.xlarge",
            predictor_num_threads=4,
            ram=16,
            gpuram=24,
            is_supported_by_a8f=True,
            aws_govcloud=False,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="g5.2xlarge",
            family="g5",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/g5/",
            hw="NVIDIA A10G",
            cores=8,
            hyperthreading=False,
            predictor_target="g5.2xlarge",
            predictor_num_threads=4,
            ram=32,
            gpuram=24,
            is_supported_by_a8f=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="g5.4xlarge",
            family="g5",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/g5/",
            hw="NVIDIA A10G",
            cores=16,
            hyperthreading=False,
            predictor_target="g5.4xlarge",
            predictor_num_threads=8,
            ram=64,
            gpuram=24,
            is_supported_by_a8f=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="g5.8xlarge",
            family="g5",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/g5/",
            hw="NVIDIA A10G",
            cores=32,
            hyperthreading=False,
            predictor_target="g5.8xlarge",
            predictor_num_threads=16,
            ram=128,
            gpuram=24,
            is_supported_by_a8f=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="g5.12xlarge",
            family="g5",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/g5/",
            hw="NVIDIA A10G",
            cores=48,
            hyperthreading=False,
            predictor_target="g5.12xlarge",
            predictor_num_threads=24,
            ram=192,
            gpuram=24 * 4,
            is_supported_by_a8f=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="g5.16xlarge",
            family="g5",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/g5/",
            hw="NVIDIA A10G",
            cores=64,
            hyperthreading=False,
            predictor_target="g5.16xlarge",
            predictor_num_threads=32,
            ram=256,
            gpuram=24,
            is_supported_by_a8f=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="g5.24xlarge",
            family="g5",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/g5/",
            hw="NVIDIA A10G",
            cores=96,
            hyperthreading=False,
            predictor_target="g5.24xlarge",
            predictor_num_threads=48,
            ram=384,
            gpuram=24 * 4,
            is_supported_by_a8f=True,
        ),
        **amazon_cloud_info_service.build_keyed_instance_description(
            label="g5.48xlarge",
            family="g5",
            instance_description_link="https://aws.amazon.com/ec2/instance-types/g5/",
            hw="NVIDIA A10G",
            cores=192,
            hyperthreading=False,
            predictor_target="g5.48xlarge",
            predictor_num_threads=96,
            ram=768,
            gpuram=24 * 8,
            is_supported_by_a8f=True,
        ),
    }


def get_supported_oracle_cloud_instances(
    oracle_cloud_info_service: CloudInfoService,
):
    return {
        "VM.Standard.A1.Flex": {
            2: InstanceDescription(
                label="VM.Standard.A1.Flex 2vCPU",
                family="VM.Standard.A1.Flex",
                instance_description_link="https://www.oracle.com/cloud/compute/arm/",
                hw="Ampere Altra",
                cores=2,
                hyperthreading=False,
                usd_per_hr=0.026,
                predictor_target="VM.Standard.A1.Flex",
                predictor_num_threads=2,
                service_provider=CloudServiceProvider.ORACLE,
                ram=None,
                gpuram=None,
                aws_govcloud=False,
            ),
            4: InstanceDescription(
                label="VM.Standard.A1.Flex 4vCPU",
                family="VM.Standard.A1.Flex",
                instance_description_link="https://www.oracle.com/cloud/compute/arm/",
                hw="Ampere Altra",
                cores=4,
                hyperthreading=False,
                usd_per_hr=0.052,
                predictor_target="VM.Standard.A1.Flex",
                predictor_num_threads=4,
                service_provider=CloudServiceProvider.ORACLE,
                ram=None,
                gpuram=None,
                aws_govcloud=False,
            ),
            8: InstanceDescription(
                label="VM.Standard.A1.Flex 8vCPU",
                family="VM.Standard.A1.Flex",
                instance_description_link="https://www.oracle.com/cloud/compute/arm/",
                hw="Ampere Altra",
                cores=8,
                hyperthreading=False,
                usd_per_hr=0.104,
                predictor_target="VM.Standard.A1.Flex",
                predictor_num_threads=8,
                service_provider=CloudServiceProvider.ORACLE,
                ram=None,
                gpuram=None,
                aws_govcloud=False,
            ),
            16: InstanceDescription(
                label="VM.Standard.A1.Flex 16vCPU",
                family="VM.Standard.A1.Flex",
                instance_description_link="https://www.oracle.com/cloud/compute/arm/",
                hw="Ampere Altra",
                cores=16,
                hyperthreading=False,
                usd_per_hr=0.208,
                predictor_target="VM.Standard.A1.Flex",
                predictor_num_threads=16,
                service_provider=CloudServiceProvider.ORACLE,
                ram=None,
                gpuram=None,
                aws_govcloud=False,
            ),
            32: InstanceDescription(
                label="VM.Standard.A1.Flex 32vCPU",
                family="VM.Standard.A1.Flex",
                instance_description_link="https://www.oracle.com/cloud/compute/arm/",
                hw="Ampere Altra",
                cores=32,
                hyperthreading=False,
                usd_per_hr=0.416,
                predictor_target="VM.Standard.A1.Flex",
                predictor_num_threads=32,
                service_provider=CloudServiceProvider.ORACLE,
                ram=None,
                gpuram=None,
                aws_govcloud=False,
            ),
            48: InstanceDescription(
                label="VM.Standard.A1.Flex 48vCPU",
                family="VM.Standard.A1.Flex",
                instance_description_link="https://www.oracle.com/cloud/compute/arm/",
                hw="Ampere Altra",
                cores=48,
                hyperthreading=False,
                usd_per_hr=0.624,
                predictor_target="VM.Standard.A1.Flex",
                predictor_num_threads=48,
                service_provider=CloudServiceProvider.ORACLE,
                ram=None,
                gpuram=None,
                aws_govcloud=False,
            ),
            64: InstanceDescription(
                label="VM.Standard.A1.Flex 64vCPU",
                family="VM.Standard.A1.Flex",
                instance_description_link="https://www.oracle.com/cloud/compute/arm/",
                hw="Ampere Altra",
                cores=64,
                hyperthreading=False,
                usd_per_hr=0.832,
                predictor_target="VM.Standard.A1.Flex",
                predictor_num_threads=64,
                service_provider=CloudServiceProvider.ORACLE,
                ram=None,
                gpuram=None,
                aws_govcloud=False,
            ),
            80: InstanceDescription(
                label="VM.Standard.A1.Flex 80vCPU",
                family="VM.Standard.A1.Flex",
                instance_description_link="https://www.oracle.com/cloud/compute/arm/",
                hw="Ampere Altra",
                cores=80,
                hyperthreading=False,
                usd_per_hr=1.040,
                predictor_target="VM.Standard.A1.Flex",
                predictor_num_threads=80,
                service_provider=CloudServiceProvider.ORACLE,
                ram=None,
                gpuram=None,
                aws_govcloud=False,
            ),
        },
    }


def get_supported_azure_cloud_instances(
    azure_cloud_info_service: CloudInfoService,
):
    return {
        # TODO: why these instances can have 8 threads support with just 2 phys cores?
        "Azure.StandardD2sv3": {
            8: InstanceDescription(
                label="Azure.StandardD2sv3",
                family="D2",
                instance_description_link="https://docs.microsoft.com/en-us/azure/virtual-machines/dv3-dsv3-series",
                hw="Intel Xeon",
                cores=2,
                hyperthreading=False,
                usd_per_hr=0.09,
                predictor_target="Azure.StandardD2sv3",
                predictor_num_threads=8,
                service_provider=CloudServiceProvider.AZURE,
                ram=8,
                gpuram=0,
                aws_govcloud=False,
            )
        },
        "Azure.StandardDS2v2": {
            8: InstanceDescription(
                label="Azure.StandardD2sv2",
                family="D2",
                instance_description_link="https://docs.microsoft.com/en-us/azure/virtual-machines/dv2-dsv2-series",
                hw="Intel Xeon",
                cores=2,
                hyperthreading=False,
                usd_per_hr=0.09,
                predictor_target="Azure.StandardDS2v2",
                predictor_num_threads=8,
                service_provider=CloudServiceProvider.AZURE,
                ram=8,
                gpuram=0,
                aws_govcloud=False,
            )
        },
        "Azure.StandardB2s": {
            8: InstanceDescription(
                label="Azure.StandardB2s",
                family="B2",
                instance_description_link="https://docs.microsoft.com/en-us/azure/virtual-machines/sizes-b-series-burstable",
                hw="Intel Xeon",
                cores=2,
                hyperthreading=False,
                usd_per_hr=0.09,
                predictor_target="Azure.StandardB2s",
                predictor_num_threads=8,
                service_provider=CloudServiceProvider.AZURE,
                ram=8,
                gpuram=0,
                aws_govcloud=False,
            )
        },
        "Azure.StandardB2ms": {
            8: InstanceDescription(
                label="Azure.StandardB2ms",
                family="B2",
                instance_description_link="https://docs.microsoft.com/en-us/azure/virtual-machines/sizes-b-series-burstable",
                hw="Intel Xeon Platinum",
                cores=2,
                hyperthreading=False,
                usd_per_hr=0.09,
                predictor_target="Azure.StandardB2ms",
                predictor_num_threads=8,
                service_provider=CloudServiceProvider.AZURE,
                ram=8,
                gpuram=0,
                aws_govcloud=False,
            )
        },
        "Azure.StandardD2sv4": {
            8: InstanceDescription(
                label="Azure.StandardD2sv4",
                family="D2",
                instance_description_link="https://docs.microsoft.com/en-us/azure/virtual-machines/dv4-dsv4-series",
                hw="Intel Xeon",
                cores=2,
                hyperthreading=False,
                usd_per_hr=0.09,
                predictor_target="Azure.StandardD2sv4",
                predictor_num_threads=8,
                service_provider=CloudServiceProvider.AZURE,
                ram=8,
                gpuram=0,
                aws_govcloud=False,
            )
        },
        "Azure.StandardD2sv5": {
            8: InstanceDescription(
                label="Azure.StandardD2sv5",
                family="D2",
                instance_description_link="https://docs.microsoft.com/en-us/azure/virtual-machines/dv5-dsv5-series",
                hw="Intel Xeon Platinum Ice Lake",
                cores=2,
                hyperthreading=False,
                usd_per_hr=0.09,
                predictor_target="Azure.StandardD2sv5",
                predictor_num_threads=8,
                service_provider=CloudServiceProvider.AZURE,
                ram=8,
                gpuram=0,
                aws_govcloud=False,
            )
        },
        "Azure.StandardDS1v2": {
            2: InstanceDescription(
                label="Azure.StandardD1sv2",
                family="D1",
                instance_description_link="https://docs.microsoft.com/en-us/azure/virtual-machines/dv2-dsv2-series",
                hw="Intel Xeon",
                cores=1,
                hyperthreading=False,
                usd_per_hr=0.073,
                predictor_target="Azure.StandardDS1v2",  # TODO: typo?
                predictor_num_threads=2,
                service_provider=CloudServiceProvider.AZURE,
                ram=3.5,
                gpuram=0,
                aws_govcloud=False,
            )
        },
        "Azure.StandardB4ms": {
            8: InstanceDescription(
                label="Azure.StandardB4ms",
                family="B4",
                instance_description_link="https://docs.microsoft.com/en-us/azure/virtual-machines/sizes-b-series-burstable",
                hw="Intel Xeon Platinum",
                cores=4,
                hyperthreading=False,
                usd_per_hr=0.166,
                predictor_target="Azure.StandardB4ms",
                predictor_num_threads=8,
                service_provider=CloudServiceProvider.AZURE,
                ram=16,
                gpuram=0,
                aws_govcloud=False,
            )
        },
        "Azure.StandardE2sv3": {
            8: InstanceDescription(
                label="Azure.StandardE2sv3",
                family="E2",
                instance_description_link="https://docs.microsoft.com/en-us/azure/virtual-machines/ev3-esv3-series",
                hw="Intel Xeon Platinum",
                cores=2,
                hyperthreading=False,
                usd_per_hr=0.126,
                predictor_target="Azure.StandardE2sv3",
                predictor_num_threads=8,
                service_provider=CloudServiceProvider.AZURE,
                ram=16,
                gpuram=0,
                aws_govcloud=False,
            )
        },
        "Azure.StandardF2sv2": {
            8: InstanceDescription(
                label="Azure.StandardF2sv2",
                family="F2",
                instance_description_link="https://docs.microsoft.com/en-us/azure/virtual-machines/fsv2-series",
                hw="Intel Xeon Platinum",
                cores=2,
                hyperthreading=False,
                usd_per_hr=0.0846,
                predictor_target="Azure.StandardF2sv2",
                predictor_num_threads=8,
                service_provider=CloudServiceProvider.AZURE,
                ram=4,
                gpuram=0,
                aws_govcloud=False,
            )
        },
        "Azure.StandardDS11-1v2": {
            4: InstanceDescription(
                label="Azure.StandardDS11-1v2",
                family="DS11",
                instance_description_link="https://docs.microsoft.com/en-us/azure/virtual-machines/constrained-vcpu",
                hw="Intel Xeon",
                cores=1,
                hyperthreading=False,
                usd_per_hr=0.185,
                predictor_target="Azure.StandardDS11-1v2",
                predictor_num_threads=4,
                service_provider=CloudServiceProvider.AZURE,
                ram=14,
                gpuram=0,
                aws_govcloud=False,
            )
        },
        "Azure.StandardDS11v2": {
            4: InstanceDescription(
                label="Azure.StandardDS11v2",
                family="DS11",
                instance_description_link="https://docs.microsoft.com/en-us/azure/virtual-machines/dv2-dsv2-series-memory",
                hw="Intel Xeon",
                cores=2,
                hyperthreading=False,
                usd_per_hr=0.185,
                predictor_target="Azure.StandardDS11v2",
                predictor_num_threads=4,
                service_provider=CloudServiceProvider.AZURE,
                ram=14,
                gpuram=0,
                aws_govcloud=False,
            )
        },
        "Azure.StandardE2s_v4": {
            4: InstanceDescription(
                label="Azure.StandardE2sv4",
                family="E2",
                instance_description_link="https://docs.microsoft.com/en-us/azure/virtual-machines/ev4-esv4-series",
                hw="Intel Xeon Platinum",
                cores=2,
                hyperthreading=False,
                usd_per_hr=0.126,
                predictor_target="Azure.StandardE2s_v4",  # TODO inconsistent
                predictor_num_threads=4,
                service_provider=CloudServiceProvider.AZURE,
                ram=16,
                gpuram=0,
                aws_govcloud=False,
            )
        },
        "Azure.StandardE2dsv4": {
            4: InstanceDescription(
                label="Azure.StandardE2dsv4",
                family="E2",
                instance_description_link="https://docs.microsoft.com/en-us/azure/virtual-machines/edv4-edsv4-series",
                hw="Intel Xeon Platinum",
                cores=2,
                hyperthreading=False,
                usd_per_hr=0.144,
                predictor_target="Azure.StandardE2dsv4",
                predictor_num_threads=4,
                service_provider=CloudServiceProvider.AZURE,
                ram=16,
                gpuram=0,
                aws_govcloud=False,
            )
        },
        "Azure.StandardE2asv4": {
            4: InstanceDescription(
                label="Azure.StandardE2asv4",
                family="E2",
                instance_description_link="https://docs.microsoft.com/en-us/azure/virtual-machines/eav4-easv4-series",
                hw="Intel Xeon Platinum",
                cores=2,
                hyperthreading=False,
                usd_per_hr=0.126,
                predictor_target="Azure.StandardE2asv4",
                predictor_num_threads=4,
                service_provider=CloudServiceProvider.AZURE,
                ram=16,
                gpuram=0,
                aws_govcloud=False,
            )
        },
        "Azure.StandardE2sv5": {
            4: InstanceDescription(
                label="Azure.StandardE2sv5",
                family="E2",
                instance_description_link="https://docs.microsoft.com/en-us/azure/virtual-machines/ev5-esv5-series",
                hw="Intel Xeon Platinum Ice Lake",
                cores=2,
                hyperthreading=False,
                usd_per_hr=0.126,
                predictor_target="Azure.StandardE2sv5",
                predictor_num_threads=4,
                service_provider=CloudServiceProvider.AZURE,
                ram=16,
                gpuram=0,
                aws_govcloud=False,
            )
        },
        "Azure.StandardE2dsv5": {
            4: InstanceDescription(
                label="Azure.StandardE2dsv5",
                family="E2",
                instance_description_link="https://docs.microsoft.com/en-us/azure/virtual-machines/edv5-edsv5-series",
                hw="Intel Xeon Platinum Ice Lake",
                cores=2,
                hyperthreading=False,
                usd_per_hr=0.144,
                predictor_target="Azure.StandardE2dsv5",
                predictor_num_threads=4,
                service_provider=CloudServiceProvider.AZURE,
                ram=16,
                gpuram=0,
                aws_govcloud=False,
            )
        },
        "Azure.StandardE2asv5": {
            4: InstanceDescription(
                label="Azure.StandardE2asv5",
                family="E2",
                instance_description_link="https://docs.microsoft.com/en-us/azure/virtual-machines/easv5-eadsv5-series",
                hw="Intel Xeon Platinum Ice Lake",
                cores=2,
                hyperthreading=False,
                usd_per_hr=0.1129,
                predictor_target="Azure.StandardE2asv5",
                predictor_num_threads=4,
                service_provider=CloudServiceProvider.AZURE,
                ram=16,
                gpuram=0,
                aws_govcloud=False,
            )
        },
        "Azure.StandardE2adsv5": {
            4: InstanceDescription(
                label="Azure.StandardE2adsv5",
                family="E2",
                instance_description_link="https://docs.microsoft.com/en-us/azure/virtual-machines/easv5-eadsv5-series",
                hw="Intel Xeon Platinum Ice Lake",
                cores=2,
                hyperthreading=False,
                usd_per_hr=0.1309,
                predictor_target="Azure.StandardE2adsv5",
                predictor_num_threads=4,
                service_provider=CloudServiceProvider.AZURE,
                ram=16,
                gpuram=0,
                aws_govcloud=False,
            )
        },
        "Azure.StandardD2dsv4": {
            4: InstanceDescription(
                label="Azure.StandardD2dsv4",
                family="D2",
                instance_description_link="https://docs.microsoft.com/en-us/azure/virtual-machines/ddv4-ddsv4-series",
                hw="Intel Xeon Platinum",
                cores=2,
                hyperthreading=False,
                usd_per_hr=0.1129,
                predictor_target="Azure.StandardD2dsv4",
                predictor_num_threads=4,
                service_provider=CloudServiceProvider.AZURE,
                ram=8,
                gpuram=0,
                aws_govcloud=False,
            )
        },
        "Azure.StandardD2asv4": {
            4: InstanceDescription(
                label="Azure.StandardD2asv4",
                family="D2",
                instance_description_link="https://docs.microsoft.com/en-us/azure/virtual-machines/dav4-dasv4-series",
                hw="Intel Xeon Platinum",
                cores=2,
                hyperthreading=False,
                usd_per_hr=0.096,
                predictor_target="Azure.StandardD2asv4",
                predictor_num_threads=4,
                service_provider=CloudServiceProvider.AZURE,
                ram=8,
                gpuram=0,
                aws_govcloud=False,
            )
        },
        "Azure.StandardD2dsv5": {
            4: InstanceDescription(
                label="Azure.StandardD2dsv5",
                family="D2",
                instance_description_link="https://docs.microsoft.com/en-us/azure/virtual-machines/ddv5-ddsv5-series",
                hw="Intel Xeon Platinum Ice Lake",
                cores=2,
                hyperthreading=False,
                usd_per_hr=0.1129,
                predictor_target="Azure.StandardD2dsv5",
                predictor_num_threads=4,
                service_provider=CloudServiceProvider.AZURE,
                ram=8,
                gpuram=0,
                aws_govcloud=False,
            )
        },
        "Azure.StandardD2asv5": {
            4: InstanceDescription(
                label="Azure.StandardD2asv5",
                family="D2",
                instance_description_link="https://docs.microsoft.com/en-us/azure/virtual-machines/dasv5-dadsv5-series",
                hw="Intel Xeon Platinum Ice Lake",
                cores=2,
                hyperthreading=False,
                usd_per_hr=0.086,
                predictor_target="Azure.StandardD2asv5",
                predictor_num_threads=4,
                service_provider=CloudServiceProvider.AZURE,
                ram=8,
                gpuram=0,
                aws_govcloud=False,
            )
        },
        "Azure.StandardD2adsv5": {
            4: InstanceDescription(
                label="Azure.StandardD2adsv5",
                family="D2",
                instance_description_link="https://docs.microsoft.com/en-us/azure/virtual-machines/dasv5-dadsv5-series",
                hw="Intel Xeon Platinum Ice Lake",
                cores=2,
                hyperthreading=False,
                usd_per_hr=0.103,
                predictor_target="Azure.StandardD2adsv5",
                predictor_num_threads=4,
                service_provider=CloudServiceProvider.AZURE,
                ram=8,
                gpuram=0,
                aws_govcloud=False,
            )
        },
    }


def get_supported_google_cloud_instances(
    google_cloud_info_service: CloudInfoService,
):
    return {
        # TODO: rename all predictor targets here to label values? then change 'label' used in Rafal's/GCP code back to 'predictor_target'
        **google_cloud_info_service.build_keyed_instance_description(
            label="c2-standard-4",
            family="C2",
            instance_description_link="https://cloud.google.com/compute/docs/compute-optimized-machines#c2_machine_types",
            hw="Intel Cascade Lake CPU",
            cores=4,
            hyperthreading=False,
            predictor_target="gcp.C2",
            predictor_num_threads=8,
            ram=16,
            gpuram=0,
        ),
        **google_cloud_info_service.build_keyed_instance_description(
            label="c2-standard-8",
            family="C2",
            instance_description_link="https://cloud.google.com/compute/docs/compute-optimized-machines#c2_machine_types",
            hw="Intel Cascade Lake CPU",
            cores=8,
            hyperthreading=False,
            predictor_target="gcp.C2_8cores",
            predictor_num_threads=16,
            ram=32,
            gpuram=0,
        ),
        **google_cloud_info_service.build_keyed_instance_description(
            label="e2-standard-2",
            family="E2",
            instance_description_link="https://cloud.google.com/compute/docs/general-purpose-machines#e2_machine_types",
            hw="Intel Skylake CPU",
            cores=2,
            hyperthreading=False,
            predictor_target="gcp.E2_2cores_8gb",
            predictor_num_threads=4,
            ram=8,
            gpuram=0,
        ),
        **google_cloud_info_service.build_keyed_instance_description(
            label="e2-standard-4",
            family="E2",
            instance_description_link="https://cloud.google.com/compute/docs/general-purpose-machines#e2_machine_types",
            hw="Intel Skylake CPU",
            cores=4,
            hyperthreading=False,
            predictor_target="gcp.e2_4cores",
            predictor_num_threads=8,
            ram=16,
            gpuram=0,
        ),
        **google_cloud_info_service.build_keyed_instance_description(
            label="e2-standard-8",
            family="E2",
            instance_description_link="https://cloud.google.com/compute/docs/general-purpose-machines#e2_machine_types",
            hw="Intel Skylake CPU",
            cores=8,
            hyperthreading=False,
            predictor_target="gcp.e2_8cores",
            predictor_num_threads=16,
            ram=32,
            gpuram=0,
        ),
        **google_cloud_info_service.build_keyed_instance_description(
            label="e2-standard-16",
            family="E2",
            instance_description_link="https://cloud.google.com/compute/docs/general-purpose-machines#e2_machine_types",
            hw="Intel Skylake CPU",
            cores=16,
            hyperthreading=False,
            predictor_target="gcp.e2_16cores",
            predictor_num_threads=32,
            ram=64,
            gpuram=0,
        ),
    }



import configparser
import os
from pathlib import Path

import appdirs

from autumn8.exceptions import UserActionRequiredException

APP_NAME = "autumn8"
APP_AUTHOR = "autumn8"


def store_api_creds(user_id, api_key):
    app_config_dir = appdirs.user_config_dir(APP_NAME, APP_AUTHOR)
    app_config_path = os.path.join(app_config_dir, "autumn8.ini")
    print(f"Saving configuration under {app_config_path}")

    config = configparser.ConfigParser()
    config.read([app_config_path])
    if not config.has_section("api_access"):
        config.add_section("api_access")
    config.set("api_access", "user_id", user_id)
    config.set("api_access", "api_key", api_key)

    path = Path(app_config_dir)
    path.mkdir(parents=True, exist_ok=True)
    file = open(app_config_path, "w")
    config.write(file)


def retrieve_api_creds():
    try:
        app_config_dir = appdirs.user_config_dir(APP_NAME, APP_AUTHOR)
        app_config_path = os.path.join(app_config_dir, "autumn8.ini")
        config = configparser.ConfigParser()
        config.read(app_config_path)
        if not config.has_section("api_access"):
            config.add_section("api_access")

        user_id = config.get("api_access", "user_id")
        api_key = config.get("api_access", "api_key")
        return user_id, api_key
    # TODO: pick a more specific set of exceptions here
    except Exception as exc:
        raise UserActionRequiredException(
            f"API key is missing! To configure API access, please visit your profile page and generate an API key, then run `autumn8-cli login`"
        ) from exc



# this is a wrapper for the `asyncio` module
# pyright: reportWildcardImportFromLibrary=false
# pylint: disable=wildcard-import,unused-wildcard-import

import asyncio
import contextvars
import functools
from asyncio import *
from typing import Any


# https://stackoverflow.com/a/69165563
# TODO: replace with asyncio.to_thread once Python gets updated to 3.9+
async def to_thread(func, /, *args: Any, **kwargs: Any):
    loop = asyncio.get_running_loop()
    ctx = contextvars.copy_context()
    func_call = functools.partial(ctx.run, func, *args, **kwargs)
    return await loop.run_in_executor(None, func_call)



import os
import sys
import unittest
from unittest.mock import MagicMock, patch

sys.path.append(os.path.join(os.path.dirname(__file__), ".."))

from autumn8.common.config.supported_instances import (
    find_instance_config,
    get_supported_cloud_instances,
)


def mock_cloud_info_service_methods(ServiceMock):
    ServiceMock.get_instance_info = MagicMock(
        side_effect=(
            lambda instance_label: {
                "type": instance_label,
                "onDemandPrice": 123,
            }
        )
    )

    ServiceMock.get_instance_pricing = MagicMock(return_value=123)

    ServiceMock.get_cloud_products_info = MagicMock(
        return_value=[{"onDemandPrice": 123}]
    )


class TestCloudInstances(unittest.TestCase):
    @patch(
        "autumn8.common.config.supported_instances.CloudInfoService.get_instance_pricing"
    )
    def test_instance_descriptions_constraints(self, mock_get_instance_pricing):
        """
        Test that get_supported_cloud_instances methods return valid InstanceDescriptions
        """
        mock_get_instance_pricing.return_value = 123

        for (
            predictor_target_key,
            instances_by_num_threads,
        ) in get_supported_cloud_instances().items():
            for (
                num_threads_key,
                instance,
            ) in instances_by_num_threads.items():
                self.assertEqual(
                    predictor_target_key, instance.predictor_target
                )
                self.assertTrue(
                    instance.instance_description_link,
                    msg="The instance description link must not be empty",
                )

    @unittest.skip(
        "Instance descriptions have their core/thread counts all over the place"
    )
    @patch(
        "autumn8.common.config.supported_instances.CloudInfoService.get_instance_pricing"
    )
    def test_instance_descriptions_core_counts(self, mock_get_instance_pricing):
        """
        Test that get_supported_cloud_instances have proper number of threads
        """
        mock_get_instance_pricing.return_value = 123

        for (
            instances_by_num_threads
        ) in get_supported_cloud_instances().values():
            for instance in instances_by_num_threads.values():
                instance_name = instance.predictor_target
                cores = instance.cores
                threads = instance.predictor_num_threads
                hyperthreading = instance.hyperthreading
                self.assertTrue(
                    # FIXME: WTF, sometimes cores are threads and threads are cores? why??????? WHYYYYWYYWYYYY!?!?!?!?! ffs
                    threads == cores if hyperthreading else cores / 2,
                    msg="The number of threads (%s) on %s must be properly related to the %s cores it has with hyperthreading=%s"
                    % (threads, instance_name, cores, hyperthreading),
                )

    @patch(
        "autumn8.common.config.supported_instances.CloudInfoService.get_instance_pricing"
    )
    def test_find_instance_config_finds_proper_instance(
        self, mock_get_instance_pricing
    ):
        """
        Test that find_instance_config can find example instances
        """
        mock_get_instance_pricing.return_value = 123

        instances_to_check = [
            "c5.large",
            "Azure.StandardE2dsv4",
            "gcp.e2_16cores",
        ]

        for predictor_target in instances_to_check:
            instance = find_instance_config(
                predictor_target, fetch_data_from_cloud_info=False
            )

            self.assertEqual(instance.predictor_target, predictor_target)


if __name__ == "__main__":
    unittest.main()



import asyncio
import base64
import json
import os
import posixpath
import uuid
from typing import Dict, List, Optional, Union

import click
import httpx
import questionary
from questionary import Choice

import autumn8.lib.api.cloud as cloud_api
from autumn8.cli import options
from autumn8.cli.cli_environment import CliEnvironment
from autumn8.cli.interactive import (
    coro_click_command,
    get_deployment,
    normalize_args,
    validate_and_ask_about_docker_image_name,
    validate_and_ask_about_schedule,
)
from autumn8.common.config.settings import CloudServiceProvider
from autumn8.lib import api, logging

logger = logging.getLogger(__name__)


@click.group(context_settings={"token_normalize_func": normalize_args})
def cloud_commands_group():
    pass


@cloud_commands_group.command()
@options.use_environment
@options.use_organization_id
@options.use_quiet_mode
@options.use_cloud_provider_picker(optional=True)
@click.option(
    "-m",
    "--model_id",
    help="Model ID to get the deployments for",
    prompt_required=False,
    default=None,
)
def list_deployments(
    organization_id: int,
    model_id: int,
    environment: CliEnvironment,
    cloud_provider: CloudServiceProvider,
    quiet,
):
    """List running deployments."""
    logger.info("Fetching the list of deployments...")
    deployments = cloud_api.get_running_deployments(
        organization_id,
        environment,
        model_id=model_id,
        service_provider=cloud_provider,
    )

    click.echo(json.dumps([depl.dict() for depl in deployments], indent=4))
    return


@cloud_commands_group.command()
@click.option(
    "-hw",
    "-t",
    "--machine_type",
    type=str,
    help="Server type to use for the deployment",
    # TODO: add a better interactive prompt listing all available servers
)
@options.use_environment
@options.use_organization_id
@options.use_quiet_mode
@click.option(
    "-m",
    "--model_id",
    prompt=True,
    type=int,
    help="Model ID to deploy",
    # TODO: add a better interactive prompt listing all available models
)
@click.option(
    "-s/-i",
    "should_schedule",
    "--schedule/--immediate",
    is_flag=True,
    default=False,
    help="Schedule the deployment to run in the future",
)
@click.option(
    "--schedule_on",
    type=str,
    help="Schedule the deployment on given date",
)
@click.option(
    "--deployment_id",
    type=str,
    help="Update an existing deployment, retaining its URL",
)
@options.use_cloud_provider_picker(
    # prompt disabled, as only A8F works atm
    optional=True,
    default_value=CloudServiceProvider.AUTUMN8,
)
def deploy(
    organization_id: int,
    model_id: int,
    should_schedule: bool,
    schedule_on: Optional[str],
    deployment_id: Optional[str],
    machine_type: Optional[str],
    environment: CliEnvironment,
    cloud_provider: CloudServiceProvider,
    quiet,
):
    """Deploy a model from AutoDL onto cloud."""

    if machine_type is None:
        machine_type = questionary.text(
            message="Machine type (ie. c5.2xlarge)"
        ).unsafe_ask()
    if machine_type is None:
        raise RuntimeError("Machine type was not picked, aborting...")

    schedule_on = validate_and_ask_about_schedule(should_schedule, schedule_on)

    logger.info(
        "Launching a new deployment with %s...",
        machine_type,
    )
    deployments = cloud_api.deploy(
        organization_id,
        environment,
        machine_type=machine_type,
        service_provider=cloud_provider,
        schedule_on=schedule_on,
        deployment_id=deployment_id,
        model_id=model_id,
    )

    click.echo(json.dumps(deployments, indent=4))


@cloud_commands_group.command()
@options.use_environment
@options.use_organization_id
@options.use_quiet_mode
@options.use_cloud_provider_picker(
    # prompt disabled, as only A8F works atm
    optional=True,
    default_value=CloudServiceProvider.AUTUMN8,
)
@click.option(
    "-d",
    "--deployment_id",
    prompt=True,
    help="ID of the deployment to terminate",
)
def terminate_deployment(
    organization_id: int,
    deployment_id: str,
    environment: CliEnvironment,
    cloud_provider: CloudServiceProvider,
    quiet,
):
    """Terminate a running deployment."""
    response = cloud_api.terminate_deployment(
        organization_id, environment, deployment_id, cloud_provider
    )
    click.echo(json.dumps(response, indent=4))


DEFAULT_DOCKER_CONTAINER_REGISTRY = "docker.io"


def build_docker_processing_query_string(
    organization_id: int, machine_type: str
) -> str:
    # TODO: drop obsolete, unused query string?
    container_id = str(uuid.uuid4())

    container_image_name = ""
    deployment_id = 0
    memory = 0

    return "+".join(
        [
            container_id,
            str(organization_id),
            container_image_name,
            str(memory),
            str(deployment_id),
            machine_type,
        ]
    )


@cloud_commands_group.command()
@coro_click_command
@options.use_organization_id
@click.option(
    "-hw",
    "-t",
    "machine_type_param",
    "--machine_type",
    type=str,
    help="Server type to use for the deployment",
    # TODO: add a better interactive prompt listing all available servers
)
@options.use_environment
@options.use_quiet_mode
@click.option(
    "-i",
    "--docker_image_name",
    default="",
    help="Docker image to be used for inference",
)
@click.option(
    "-p",
    "--docker_port",
    type=int,
    default=80,
    prompt=True,
    help="Docker container port that handles HTTP requests",
)
@click.option(
    "-r",
    "--docker_container_registry",
    type=str,
    default=DEFAULT_DOCKER_CONTAINER_REGISTRY,
    help="Docker registry to use the docker image from",
)
@click.option(
    "-h",
    "--container_http_request_path",
    type=str,
    default="",
    prompt=True,
    help="HTTP path on the Docker container to query against",
)
@click.option(
    "-H",
    "--custom_headers",
    type=str,
    multiple=True,
    default=[],
    help="Custom HTTP headers to pass",
)
@click.option(
    "-I",
    "--input",
    type=str,
    help="JSON / Raw Text HTTP Body input to pass within the HTTP request",
)
async def run_docker(
    organization_id: int,
    machine_type_param: Optional[str],
    environment: CliEnvironment,
    quiet: bool,
    docker_image_name: str = "",
    docker_port: Optional[int] = None,
    docker_container_registry: Optional[
        str
    ] = DEFAULT_DOCKER_CONTAINER_REGISTRY,
    container_http_request_path: str = "",
    custom_headers: List[str] = [],
    # disabled, cause these are not working properly right now on a8f
    # docker_environment: Optional[Dict[str, str]] = None,
    # docker_entrypoint: Optional[str] = None,
    input: Optional[str] = None,
):
    """
    Run an inference on a given Docker image by creating
    a temporary container and calling an HTTP request against it
    """
    machine_type: str = (
        questionary.text(message="Machine type (ie. c5.2xlarge)").unsafe_ask()
        if machine_type_param is None
        else machine_type_param
    )
    # AWS-only, quick hack
    machine_type = machine_type.replace(".", "-")

    (
        docker_container_registry_from_image_name,
        docker_image_name,
    ) = await validate_and_ask_about_docker_image_name(docker_image_name)

    if docker_container_registry_from_image_name:
        docker_container_registry = docker_container_registry_from_image_name

    if container_http_request_path.startswith("/"):
        container_http_request_path = container_http_request_path[1:]

    api_keys = api.user.fetch_org_api_keys(environment, organization_id)
    inference_api_key = api_keys[0]["api_key"]

    custom_headers_parsed = {}
    for entry in custom_headers:
        split = entry.split(":")
        if len(split) != 2:
            print(f'"{entry}" is not a valid header entry, skipping...')

        [header, value] = split
        custom_headers_parsed[header.strip()] = value.strip()

    async with httpx.AsyncClient(
        auth=(str(organization_id), inference_api_key)
    ) as httpClient:
        query: str = build_docker_processing_query_string(
            organization_id, machine_type
        )
        url = posixpath.join(
            environment.value.a8f_host,
            "docker-processing",
            query,
            container_http_request_path,
        )
        response: httpx.Response = await httpClient.post(
            url=url,
            json={
                "container_image_name": docker_image_name,
                "container_registry": docker_container_registry,
                "container_port": docker_port,
                "http_request_input": input,
                "http_request_path": container_http_request_path,
                "http_request_headers": custom_headers_parsed,
            },
            timeout=None,
        )

        logger.info("Container responded with:")
        click.echo(response.text)
        response.raise_for_status()


def decode_base64_model_output(model_output_base64: str):
    model_output_base64_bytes = bytes(model_output_base64, encoding="utf-8")

    model_output = base64.decodebytes(model_output_base64_bytes)
    return model_output


@cloud_commands_group.command()
@options.use_organization_id
@options.use_environment
@options.use_quiet_mode
@click.option(
    "--deployment_url", type=str, help="Public dns of your deployment"
)
@click.option("--model_id", type=int, help="Model id of model")
@click.option(
    "-I",
    "model_input",
    "--input",
    type=str,
    help="JSON / Raw Text HTTP Body input to pass within the HTTP request",
)
@click.option(
    "-f",
    "--input_file",
    type=str,
    help="filepath to file with JSON / Raw Text HTTP Body input to pass within the HTTP request",
)
@click.option(
    "-d",
    "--decode64",
    type=bool,
    is_flag=True,
    help="Decode model output from base64 string",
)
@click.option(
    "-O",
    "--output_file",
    type=str,
    help="Output file",
)
def run_inference(
    organization_id: int,
    environment: CliEnvironment,
    quiet: bool,
    deployment_url: Optional[str],
    model_id: Optional[int],
    model_input: Optional[str],
    input_file: Optional[str],
    decode64: bool,
    output_file: Optional[str],
):
    """
    Run an inference on a given deployment
    """

    api_keys = api.user.fetch_org_api_keys(environment, organization_id)
    inference_api_key = api_keys[0]["api_key"]

    if decode64 and output_file is None:
        output_file = questionary.text(
            "Please specify output file to decode base64 to",
            default="",
        ).unsafe_ask()

    if deployment_url is None:
        deployment_url = get_deployment(
            organization_id, environment, model_id=model_id
        )

    if input_file is not None:
        if os.path.exists(input_file):
            with open(input_file, "r", encoding="utf-8") as file:
                model_input = file.read()

    if deployment_url is None:
        click.echo(
            "You need to specify either deployment_url or choose existing deployment"
        )
        return

    final_input: str = (
        model_input
        if model_input is not None
        else questionary.text("JSON model input").unsafe_ask()
    )

    json_input = None
    try:
        json_input = json.loads(final_input)
    except json.JSONDecodeError:
        pass

    with httpx.Client(
        auth=(str(organization_id), inference_api_key)
    ) as httpClient:
        response: httpx.Response = httpClient.post(
            url=deployment_url,
            json=json_input if json_input is not None else None,
            content=final_input if json_input is None else None,
            timeout=None,
        )
        if decode64:
            data = json.loads(response.text)
            if "message" in data and "output" in data["message"]:
                model_output_base64 = data["message"]["output"]
                model_output = decode_base64_model_output(model_output_base64)
                if (
                    output_file is None
                ):  # check for this is above, additional check for pylint
                    return
                with open(output_file, "wb") as f:
                    f.write(model_output)
            else:
                logger.info("Deployment responded with:")
                click.echo(response.text)
                response.raise_for_status()
        else:
            logger.info("Deployment responded with:")
            click.echo(response.text)
            response.raise_for_status()
            if output_file is not None:
                with open(output_file, "w") as f:
                    f.write(response.text)



import enum
from dataclasses import dataclass
from typing import Optional


@dataclass
class EnvironmentProperties:
    app_host: str
    s3_host: str
    a8f_host: str
    s3_bucket_name: str
    s3_bucket_root_folder: Optional[str]


# TODO remove - official CLI builds should only point toward production


class CliEnvironment(enum.Enum):
    LOCALHOST = EnvironmentProperties(
        app_host="http://localhost",
        s3_host="http://localhost:4566",
        a8f_host="http://localhost:7250",
        s3_bucket_name="predictor-bucket",
        s3_bucket_root_folder=None,
    )
    STAGING = EnvironmentProperties(
        app_host="http://staging.autumn8.ai",
        s3_host="https://s3-accelerate.amazonaws.com",
        a8f_host="http://autumn8functions.default.aws.staging.autumn8.ai",
        s3_bucket_name="autodl-staging",
        s3_bucket_root_folder="autodl-staging",
    )
    PRODUCTION = EnvironmentProperties(
        app_host="https://autodl.autumn8.ai",
        s3_host="https://s3-accelerate.amazonaws.com",
        a8f_host="https://autumn8functions.default.aws.autumn8.ai",
        s3_bucket_name="autodl-staging",
        s3_bucket_root_folder="autodl-production",
    )



import urllib.parse
from http import HTTPStatus
from typing import Dict, Union

import httpx
import requests


def url_with_params(url: str, params: Dict[str, Union[str, int, None]]) -> str:
    params_skipping_none = dict(
        (k, v) for k, v in params.items() if v is not None
    )
    url_parse = urllib.parse.urlparse(url)
    url_new_query = urllib.parse.urlencode(params_skipping_none)
    url_parse = url_parse._replace(query=url_new_query)

    new_url = urllib.parse.urlunparse(url_parse)
    return str(new_url)


def require_ok_response(
    response: Union[requests.Response, httpx.Response]
) -> None:
    if response.status_code == HTTPStatus.UNAUTHORIZED:
        raise Exception(
            f"Received response {response.status_code}:\n{response.text}\n\nUser not authenticated; please run `autumn8-cli login` to authorize your CLI"
        )
    if response.status_code != HTTPStatus.OK:
        raise Exception(
            f"Received response {response.status_code}:\n{response.text}"
        )



import enum
from typing import Dict

from autumn8.common.config.settings import CloudServiceProvider

ZoneConfig = Dict[CloudServiceProvider, str]

DEFAULT_CLOUD_ZONES_CONFIG: ZoneConfig = {
    CloudServiceProvider.AMAZON: "us-east-1",
    CloudServiceProvider.GOOGLE: "us-central1",
    CloudServiceProvider.ORACLE: "us-sanjose-1",
    CloudServiceProvider.AZURE: "useast",
}


class Sla(str, enum.Enum):
    LATENCY = "latency"
    THROUGHPUT = "throughput"
    COST_PERFORMANCE = "cost_performance"
    TOTAL_ENERGY = "total_energy"
    EMISSIONS = "emissions"



import io
import json
import os
from threading import Lock
from typing import Any, Dict, List, Literal, Optional, TypedDict, Union

import appdirs
import boto3
import httpx
import requests
from mypy_boto3_s3 import S3Client, S3ServiceResource
from mypy_boto3_s3.type_defs import PartTypeDef
from mypy_boto3_sts.type_defs import CredentialsTypeDef
from requests.auth import HTTPBasicAuth
from tqdm.contrib.concurrent import thread_map

from autumn8.cli import pending_uploads
from autumn8.cli.analyze import is_model_file_link_external, s3path_join
from autumn8.cli.cli_environment import CliEnvironment
from autumn8.lib import logging
from autumn8.lib.api_creds import retrieve_api_creds
from autumn8.lib.http import require_ok_response, url_with_params
from autumn8.types.router import NewModelInfo

DEFAULT_MAX_UPLOAD_WORKERS = 4

APP_NAME = "autumn8"
APP_AUTHOR = "autumn8"

logger = logging.getLogger(__name__)

data_dir = appdirs.user_data_dir(APP_NAME, APP_AUTHOR)


def fetch_user_data(environment: CliEnvironment):
    user_id, api_key = None, None
    autodl_host = environment.value.app_host
    try:
        user_id, api_key = retrieve_api_creds()
    except:
        raise Exception(
            f"API key is missing! To configure API access, please visit {autodl_host}/profile and generate an API key, then run `autumn8-cli login`"
        )

    user_api_route = f"{autodl_host}/api/user"
    response = requests.get(
        user_api_route,
        headers={"Content-Type": "application/json"},
        auth=HTTPBasicAuth(user_id, api_key),
    )

    require_ok_response(response)
    return json.loads(response.text)["user"]


def get_model(environment: CliEnvironment, organization_id: int, model_id: int):
    autodl_host = environment.value.app_host
    api_route = f"{autodl_host}/api/lab/model/stub"
    logger.info("Fetching model with id=%s", model_id)
    response = requests.get(
        url_with_params(
            api_route,
            {"organization_id": organization_id, "model_id": model_id},
        ),
        headers={"Content-Type": "application/json"},
        auth=HTTPBasicAuth(*retrieve_api_creds()),
    )

    require_ok_response(response)
    return response.json()


def delete_model(
    environment: CliEnvironment,
    organization_id: int,
    model_id: int,
):
    autodl_host = environment.value.app_host
    api_route = f"{autodl_host}/api/lab/model"
    logger.info("Deleting model with id=%s", model_id)
    response = requests.delete(
        url_with_params(
            api_route,
            {"organization_id": organization_id, "model_id": model_id},
        ),
        headers={"Content-Type": "application/json"},
        auth=HTTPBasicAuth(*retrieve_api_creds()),
    )

    require_ok_response(response)
    return response.json()


def post_model(
    environment: CliEnvironment,
    organization_id: int,
    model_config: NewModelInfo,
):
    autodl_host = environment.value.app_host
    api_route = f"{autodl_host}/api/lab/model"
    logger.info("Submitting model to %s", api_route)
    response = requests.post(
        url_with_params(api_route, {"organization_id": organization_id}),
        headers={"Content-Type": "application/json"},
        data=json.dumps(model_config.dict()),
        auth=HTTPBasicAuth(*retrieve_api_creds()),
    )

    require_ok_response(response)
    return response.json()


FileUploadFileType = Union[Literal["model"], Literal["input"]]


def normal_or_multipart_upload(
    organization_id: int,
    environment: CliEnvironment,
    file: io.BufferedReader,
    file_name: str,
    file_type: FileUploadFileType,
    resume_args,
    id_key,
    mpu_id=None,
    max_upload_workers=DEFAULT_MAX_UPLOAD_WORKERS,
):
    file.seek(0, 2)  # seek to end of file
    total_bytes = file.tell()
    file.seek(0)

    permissions = request_temporary_model_file_upload_permissions(
        environment, organization_id, file_name, file_type
    )

    # AWS dissallow multipart upload of files under 5MB
    if total_bytes < 6 * 1024**2:
        return normal_upload(environment, permissions, file)
    else:
        return multipart_upload(
            environment=environment,
            permissions=permissions,
            file=file,
            resume_args=resume_args,
            id_key=id_key,
            mpu_id=mpu_id,
            max_upload_workers=max_upload_workers,
        )


class FileUploadPermissionsResponse(TypedDict):
    message: str
    object_key: str
    bucket_name: str
    credentials: CredentialsTypeDef


def request_temporary_model_file_upload_permissions(
    environment: CliEnvironment,
    organization_id: int,
    file_name: str,
    file_type: FileUploadFileType,
):
    autodl_host = environment.value.app_host
    api_route = f"{autodl_host}/api/lab/model/upload_sts"
    response = requests.get(
        url_with_params(
            api_route,
            {
                "organization_id": organization_id,
                "file_name": file_name,
                "file_type": file_type,
            },
        ),
        headers={"Content-Type": "application/json"},
        auth=HTTPBasicAuth(*retrieve_api_creds()),
    )

    require_ok_response(response)
    return FileUploadPermissionsResponse(response.json())


def get_s3_resource_from_temporary_permissions(
    environment: CliEnvironment, permissions: FileUploadPermissionsResponse
) -> S3ServiceResource:
    credentials = permissions["credentials"]

    session = boto3.Session(
        aws_access_key_id=credentials["AccessKeyId"],
        aws_secret_access_key=credentials["SecretAccessKey"],
        aws_session_token=credentials["SessionToken"],
    )

    s3 = session.resource(
        "s3", endpoint_url=environment.value.s3_host, region_name="us-east-1"
    )

    return s3


def normal_upload(
    environment: CliEnvironment,
    permissions: FileUploadPermissionsResponse,
    file,
):
    S3 = get_s3_resource_from_temporary_permissions(environment, permissions)

    object_key = permissions["object_key"]
    S3.Bucket(permissions["bucket_name"]).Object(object_key).upload_fileobj(
        file
    )
    return object_key


def get_uploaded_parts(
    s3_client: S3Client, s3_bucket_name, s3_file_url, upload_id
) -> List[PartTypeDef]:
    res = s3_client.list_parts(
        Bucket=s3_bucket_name, Key=s3_file_url, UploadId=upload_id
    )
    return list(res["Parts"]) if "Parts" in res else []


def upload_part(
    part_number,
    s3_client: S3Client,
    s3_bucket_name,
    s3_file_url,
    file,
    part_size_in_bytes,
    total_bytes,
    mpu_id,
    parts_already_uploaded: Dict[int, PartTypeDef],
    lock,
):
    with lock:
        file.seek((part_number - 1) * part_size_in_bytes)
        data = file.read(part_size_in_bytes)

    if not len(data):
        return

    if part_number in parts_already_uploaded:
        part = parts_already_uploaded[part_number]
        # FIXME: these checks are pretty slow, I'm not sure why
        if len(data) != part.get("Size"):
            raise Exception(
                "Upload corrupted: Size mismatch: local "
                + str(len(data))
                + ", remote: "
                + str(part.get("Size"))
            )
        return
    else:
        part = s3_client.upload_part(
            Body=data,
            Bucket=s3_bucket_name,
            Key=s3_file_url,
            UploadId=mpu_id,
            PartNumber=part_number,
        )


def multipart_upload(
    environment: CliEnvironment,
    permissions: FileUploadPermissionsResponse,
    file,
    resume_args,
    id_key,
    mpu_id=None,
    max_upload_workers=DEFAULT_MAX_UPLOAD_WORKERS,
):
    file.seek(0, 2)  # seek to end of file
    total_bytes = file.tell()
    file.seek(0)  # seek to start of file
    # max total upload size is 100GB
    part_size_in_bytes = max(
        10 * 1024**2, total_bytes // 500
    )  # minimum part size on aws is 5MB, list part returns max 1000 parts

    # we have to use low-level API to be able to support
    # resumable uploads - https://docs.aws.amazon.com/AmazonS3/latest/userguide/mpu-upload-object.html
    s3_client = get_s3_resource_from_temporary_permissions(
        environment, permissions
    ).meta.client
    s3_bucket_name = permissions["bucket_name"]
    compatible_s3_file_url = permissions["object_key"]

    if mpu_id != None:
        print(f"Resuming upload with id {mpu_id}")
        parts_already_uploaded = {
            part["PartNumber"]: part
            for part in get_uploaded_parts(
                s3_client, s3_bucket_name, compatible_s3_file_url, mpu_id
            )
            if "PartNumber" in part and "Size" in part
        }
    else:
        parts_already_uploaded = {}
        mpu = s3_client.create_multipart_upload(
            Bucket=s3_bucket_name, Key=compatible_s3_file_url
        )
        mpu_id = mpu["UploadId"]
        print(f"Created new multipart upload with id {mpu_id}")

    resume_args[id_key] = mpu_id
    pending_uploads.update_upload(resume_args["run_id"], resume_args)

    # +2, because we assume there's always this final part which is smaller than the rest, and the range end has to be larger by 1 because it's not included
    part_numbers_to_upload = set(
        range(1, (total_bytes - 1) // part_size_in_bytes + 2)
    )
    lock = Lock()
    thread_map(
        lambda part_number: upload_part(
            part_number=part_number,
            s3_client=s3_client,
            s3_bucket_name=s3_bucket_name,
            s3_file_url=compatible_s3_file_url,
            file=file,
            part_size_in_bytes=part_size_in_bytes,
            parts_already_uploaded=parts_already_uploaded,
            total_bytes=total_bytes,
            mpu_id=mpu_id,
            lock=lock,
        ),
        part_numbers_to_upload,
        max_workers=max_upload_workers,
    )

    all_parts = get_uploaded_parts(
        s3_client, s3_bucket_name, compatible_s3_file_url, mpu_id
    )

    s3_client.complete_multipart_upload(
        Bucket=s3_bucket_name,
        Key=compatible_s3_file_url,
        UploadId=mpu_id,
        MultipartUpload={
            "Parts": [  # type: ignore (partial Part type works fine in practice)
                {
                    "PartNumber": part.get("PartNumber"),
                    "ETag": part.get("ETag"),
                }
                for part in all_parts
            ]
        },
    )

    return permissions["object_key"]


# TODO: fix s3 file structure so that this is not needed
def get_hacked_legacy_backwards_compatible_s3_file_url(
    s3_root_folder_name: Optional[str],
    s3_file_url: str,
):
    return s3path_join(s3_root_folder_name or "", s3_file_url)


def post_model_file(
    organization_id: int,
    environment: CliEnvironment,
    filepath_or_url: str,
    file_type: FileUploadFileType,
    resume_args,
    id_key,
    upload_id=None,
    max_upload_workers=DEFAULT_MAX_UPLOAD_WORKERS,
):
    if is_model_file_link_external(filepath_or_url):
        # attaching directly without any reuploads
        return filepath_or_url

    file_name = os.path.basename(filepath_or_url)

    with open(filepath_or_url, "rb") as file:
        return normal_or_multipart_upload(
            organization_id,
            environment=environment,
            file=file,
            file_name=file_name,
            file_type=file_type,
            resume_args=resume_args,
            id_key=id_key,
            mpu_id=upload_id,
            max_upload_workers=max_upload_workers,
        )


def async_prediction(
    environment: CliEnvironment, organization_id: int, model_id: int
):
    autodl_host = environment.value.app_host
    new_url = url_with_params(
        f"{autodl_host}/api/lab/model/async_prediction",
        {
            "model_id": model_id,
            "organization_id": organization_id,
        },
    )
    response = requests.post(
        new_url,
        auth=HTTPBasicAuth(*retrieve_api_creds()),
    )
    require_ok_response(response)
    return response



import os
import pickle
from pathlib import Path

import appdirs
import click

from autumn8.cli.cli_environment import CliEnvironment

APP_NAME = "autumn8"
APP_AUTHOR = "autumn8"

data_dir = appdirs.user_data_dir(APP_NAME, APP_AUTHOR)
RESUMABLE_UPLOADS_PATH = os.path.join(data_dir, "uploads.pickle")


def retrieve_pending_uploads():
    if os.path.exists(RESUMABLE_UPLOADS_PATH):
        with open(RESUMABLE_UPLOADS_PATH, "rb") as f:
            return pickle.load(f)

    return {}


def forget_all_pending_uploads():
    os.remove(RESUMABLE_UPLOADS_PATH)


def update_upload(run_id, resume_args):
    if os.path.exists(RESUMABLE_UPLOADS_PATH):
        with open(RESUMABLE_UPLOADS_PATH, "rb") as f:
            data = pickle.load(f)

    else:
        data_path = Path(data_dir)
        data_path.mkdir(parents=True, exist_ok=True)
        data = {}

    data[run_id] = resume_args

    with open(RESUMABLE_UPLOADS_PATH, "wb") as f:
        pickle.dump(data, f)


def abort_and_forget_upload(run_id):
    resume_args = forget_upload(run_id)

    if resume_args is not None:
        environment: CliEnvironment = resume_args["environment"]
        if (
            "model_file_upload_id" in resume_args
            and resume_args["model_file_upload_id"] is not None
        ):
            abort_upload(
                environment,
                resume_args["s3_file_url"],
                resume_args["model_file_upload_id"],
            )

        if (
            "input_file_upload_id" in resume_args
            and resume_args["input_file_upload_id"] is not None
        ):
            abort_upload(
                environment,
                resume_args["s3_file_url"],
                resume_args["input_file_upload_id"],
            )


def forget_upload(run_id):
    if os.path.exists(RESUMABLE_UPLOADS_PATH):
        with open(RESUMABLE_UPLOADS_PATH, "rb") as f:
            data = pickle.load(f)

        if run_id not in data:
            return

        resume_args = data[run_id]

        data.pop(run_id)

        with open(RESUMABLE_UPLOADS_PATH, "wb") as f:
            pickle.dump(data, f)

        return resume_args

    return None


def get_mpu(environment: CliEnvironment, mpu_object_key: str, mpu_id: str):
    raise NotImplementedError()
    s3 = init_s3(environment.value.s3_host)

    s3_bucket_name = environment.value.s3_bucket_name

    mpus = list(
        s3.Bucket(s3_bucket_name).multipart_uploads.filter(
            Prefix=mpu_object_key,
        )
    )

    for m in mpus:
        if m.id == mpu_id:
            return m

    return None


def abort_upload(environment: CliEnvironment, mpu_object_key: str, mpu_id: str):
    # TODO: store permissions locally on the computer for resume, or re-request them for the same file somehow
    # https://gitlab.com/Autumn8Inc/autodl/-/issues/192
    raise NotImplementedError()
    mpu = get_mpu(environment, mpu_object_key, mpu_id)
    if mpu is None:
        click.echo(
            "The upload could not be found on S3. It may have already been aborted or completed."
        )
        return

    s3_bucket_name = environment.value.s3_bucket_name

    s3 = init_s3_client(environment.value.s3_host)

    s3.abort_multipart_upload(
        Bucket=s3_bucket_name, Key=mpu_object_key, UploadId=mpu_id
    )



import io
import os
import time
import uuid
from typing import Any, Dict, Optional

from autumn8.cli import pending_uploads
from autumn8.cli.analyze import is_model_file_link_external, s3path_join
from autumn8.cli.cli_environment import CliEnvironment
from autumn8.lib import api, logging
from autumn8.types.router import NewModelInfo

DEFAULT_MAX_UPLOAD_WORKERS = 4

logger = logging.getLogger(__name__)


def resume_upload_model(upload_task):
    raise NotImplementedError()

    try:
        return upload_model(**upload_task)
    except s3.meta.client.exceptions.NoSuchUpload:
        pending_uploads.abort_and_forget_upload(upload_task["run_id"])


def upload_model(
    environment: CliEnvironment,
    organization_id: int,
    model_config: NewModelInfo,
    model_filepath_or_url: str,
    input_file_path: Optional[str],
    max_upload_workers: int = DEFAULT_MAX_UPLOAD_WORKERS,
    model_file_upload_id: Optional[str] = None,
    input_file_upload_id: Optional[str] = None,
    run_id: Optional[str] = None,
):
    if run_id is None:  # used for resuming upload
        run_id = str(uuid.uuid4())

    function_args = locals()

    time_start = time.time()
    logger.info("Uploading the model files...")
    model_config.s3_file_url = api.lab.post_model_file(
        organization_id=organization_id,
        environment=environment,
        filepath_or_url=model_filepath_or_url,
        file_type="model",
        resume_args=function_args,
        id_key="model_file_upload_id",
        upload_id=model_file_upload_id,
        max_upload_workers=max_upload_workers,
    )
    logger.debug("Model uploaded in %.03f seconds", time.time() - time_start)

    # TODO: support uploading inputs via links
    if input_file_path != None and len(input_file_path) > 0:
        time_start = time.time()
        logger.info("Uploading the input files...")
        model_config.s3_input_file_url = api.lab.post_model_file(
            organization_id=organization_id,
            environment=environment,
            filepath_or_url=input_file_path,
            file_type="input",
            resume_args=function_args,
            id_key="input_file_upload_id",
            upload_id=input_file_upload_id,
            max_upload_workers=max_upload_workers,
        )
        logger.debug(
            "Inputs uploaded in %.03f seconds", time.time() - time_start
        )

    logger.info("Creating the model entry in AutoDL...")
    model_post_response = api.lab.post_model(
        environment, organization_id, model_config
    )
    model_id = model_post_response["id"]
    pending_uploads.forget_upload(run_id)

    logger.info("Starting up performance predictor...")
    api.lab.async_prediction(environment, organization_id, model_id)
    return model_post_response


def generate_s3_input_file_url(
    organization_id, run_id, s3_bucket_root_folder, filename
):
    if s3_bucket_root_folder is None:
        s3_bucket_root_folder = ""

    return s3path_join(
        s3_bucket_root_folder,
        "inputs",
        f"{organization_id}-{run_id}-{filename}",
    )


def generate_s3_file_url(
    organization_id,
    run_id,
    model_file_name,
    model_file,
    s3_bucket_root_folder,
    model_type,
):
    if is_model_file_link_external(model_file):
        return model_file

    additional_extension = f".{model_type}" if model_type is not None else ""
    if s3_bucket_root_folder is None:
        s3_bucket_root_folder = ""

    return s3path_join(
        s3_bucket_root_folder,
        "models",
        f"{organization_id}-{run_id}-{model_file_name}{additional_extension}",
    )



import unittest
from collections import defaultdict
from unittest import mock

from autumn8.common.config.settings import Quantization


class TestDataFileSizeCalculation(unittest.TestCase):
    @mock.patch("os.environ", defaultdict(lambda: "123456"))
    def setUp(self):
        from perf_predictor.io.bottleneck_detection import (
            calculate_data_file_size_in_kb,
        )

        self.calculate_data_file_size_in_kb = calculate_data_file_size_in_kb

    def test_fp32(self):
        self.assertEqual(
            self.calculate_data_file_size_in_kb([2, 3, 4], Quantization.FP32),
            0.096,
        )
        self.assertEqual(
            self.calculate_data_file_size_in_kb(
                [[3, 5, 1], [1, 4, 1]], Quantization.FP32
            ),
            0.076,
        )
        self.assertRaises(
            ValueError,
            self.calculate_data_file_size_in_kb,
            [2, [1, 4]],
            Quantization.FP32,
        )



import functools
import os
from pathlib import Path

SVC_ROOT = os.environ.get("SVC_ROOT", str(Path.home()) + "/onspecta_svc")


@functools.lru_cache(maxsize=None)
def app():
    class AppEnv:
        NEXT_PUBLIC_AWS_ACCOUNT_ID = os.environ["NEXT_PUBLIC_AWS_ACCOUNT_ID"]

        NEXT_PUBLIC_AWS_S3_HOST_URL = os.environ.get(
            "NEXT_PUBLIC_AWS_S3_HOST_URL"
        )

        NEXT_PUBLIC_AWS_BUCKET_NAME = os.environ["NEXT_PUBLIC_AWS_BUCKET_NAME"]

        # TODO: delete me
        NEXT_PUBLIC_FEATURE_S3_UPLOADS = True

        CONCURRENCY = (
            int(os.environ["CONCURRENCY"])
            if "CONCURRENCY" in os.environ
            else None
        )

        PORT = int(os.environ.get("PORT", 4300))

        DB_URI = os.environ.get("DB_URI")

        DB_USERNAME = os.environ["DB_USERNAME"]
        DB_PASSWORD = os.environ["DB_PASSWORD"]
        DB_ENDPOINT = os.environ["DB_ENDPOINT"]
        DB_PORT = os.environ["DB_PORT"]
        DB_DATABASE = os.environ["DB_DATABASE"]

        MODEL_DEPLOYER_AWS_ACCESS_KEY_ID = os.environ[
            "MODEL_DEPLOYER_AWS_ACCESS_KEY_ID"
        ]
        MODEL_DEPLOYER_AWS_SECRET_KEY = os.environ[
            "MODEL_DEPLOYER_AWS_SECRET_KEY"
        ]

        THIRD_PARTY_CLOUD_CLIENT_AWS_ACCESS_KEY_ID = os.environ.get(
            "THIRD_PARTY_CLOUD_CLIENT_AWS_ACCESS_KEY_ID"
        )
        THIRD_PARTY_CLOUD_CLIENT_AWS_SECRET_KEY = os.environ.get(
            "THIRD_PARTY_CLOUD_CLIENT_AWS_SECRET_KEY"
        )

        CLOUDINFO_API_PATH = os.environ["CLOUDINFO_API_PATH"]

        A8F_PROXY_BILLING_KEY = os.environ.get("A8F_PROXY_BILLING_KEY")

        API_KEYS_BUCKET = os.environ["API_KEYS_BUCKET"]
        API_KEYS_BUCKET_DIRNAME = os.environ.get("API_KEYS_BUCKET_DIRNAME")

        A8F_AWS_BUCKET_S3_HOST_URL = os.environ.get(
            "A8F_AWS_BUCKET_S3_HOST_URL", "https://s3-accelerate.amazonaws.com"
        )

        AUTODL_WORKER_FUNCTION_HOST = os.environ["AUTODL_WORKER_FUNCTION_HOST"]

        K8S_GITLAB_DOCKER_REGISTRY_ACCESS_TOKEN = os.environ.get(
            "K8S_GITLAB_DOCKER_REGISTRY_ACCESS_TOKEN"
        )

        DEPLOYMENTS_PROJECT_NAME = os.environ["DEPLOYMENTS_PROJECT_NAME"]

        AWS_EKS_DEPLOYMENTS_CLUSTER_NAME = os.environ[
            "AWS_EKS_DEPLOYMENTS_CLUSTER_NAME"
        ]
        AWS_EKS_DEPLOYMENTS_CLUSTER_URL = os.environ[
            "AWS_EKS_DEPLOYMENTS_CLUSTER_URL"
        ]
        AWS_EKS_DEPLOYMENTS_CLUSTER_DNS = os.environ[
            "AWS_EKS_DEPLOYMENTS_CLUSTER_DNS"
        ]
        AWS_EKS_IAM_ADMIN_ROLE_ARN = os.environ["AWS_EKS_IAM_ADMIN_ROLE_ARN"]

        GOOGLE_PROJECT_ID = os.environ["GOOGLE_PROJECT_ID"]
        GOOGLE_DEPLOYMENTS_CLUSTER_DNS = os.environ[
            "GOOGLE_DEPLOYMENTS_CLUSTER_DNS"
        ]
        GOOGLE_APPLICATION_CREDENTIALS = os.environ[
            "GOOGLE_APPLICATION_CREDENTIALS"
        ]

        SENDGRID_API_KEY = os.environ.get("SENDGRID_API_KEY")

        DOCKER_HOST = os.environ.get("DOCKER_HOST")

    return AppEnv


@functools.lru_cache(maxsize=None)
def cli():
    class CliEnv:
        GLOBAL_S3_HOST_URL = os.environ.get(
            "BENCHMARK_AWS_S3_HOST_URL", "https://s3-accelerate.amazonaws.com"
        )

        NEXT_PUBLIC_AWS_ACCESS_KEY_ID = os.environ[
            "NEXT_PUBLIC_AWS_ACCESS_KEY_ID"
        ]
        NEXT_PUBLIC_AWS_SECRET_KEY = os.environ["NEXT_PUBLIC_AWS_SECRET_KEY"]

    return CliEnv


@functools.lru_cache(maxsize=None)
def predictor():
    class PredictorEnv:
        GPU = os.environ["GPU"] == "1"
        TRAINING = os.environ["TRAINING"] == "1"
        MACHINE_ID = os.environ["MACHINE_ID"]
        OMP_NUM_THREADS = int(os.environ["OMP_NUM_THREADS"])
        OVERRIDE_BENCHMARK = os.environ.get("OVERRIDE_BENCHMARK") == "1"
        ARM64 = os.environ["ARM64"] == "1"

        MODELS_PATH = os.environ["MODELS_PATH"]

        TEST = os.environ.get("TEST")

        PREPARATION_MODE = int(os.environ["PREPARATION_MODE"]) == 1

    return PredictorEnv


@functools.lru_cache(maxsize=None)
def worker():
    class WorkerEnv:
        """
        Environment variables accessible both from the main app,
        and from the worker cloud function
        """

        NEXT_PUBLIC_TARGET_ENV_MODE = os.environ["NEXT_PUBLIC_TARGET_ENV_MODE"]

        NEXT_PUBLIC_AUTODL_HOST_DOMAIN = os.environ[
            "NEXT_PUBLIC_AUTODL_HOST_DOMAIN"
        ]

        DATABASE_PATH = os.environ.get(
            "DATABASE_PATH", "perf_predictor/database"
        )

        BENCHMARK_AWS_ACCESS_KEY_ID = os.environ["BENCHMARK_AWS_ACCESS_KEY_ID"]
        BENCHMARK_AWS_SECRET_KEY = os.environ["BENCHMARK_AWS_SECRET_KEY"]
        BENCHMARK_AWS_S3_HOST_URL = os.environ.get(
            "BENCHMARK_AWS_S3_HOST_URL", "https://s3-accelerate.amazonaws.com"
        )

        AUTODL_WORKER_API_TOKEN = os.environ["AUTODL_WORKER_API_TOKEN"]

    return WorkerEnv



import numpy as np
import tensorflow as tf
from tensorflow.keras.layers import Conv2D, Input

import autumn8


class MyDenseLayer(tf.keras.layers.Layer):
  def __init__(self, num_outputs):
    super(MyDenseLayer, self).__init__(name="MyDenseLayer")
    self.num_outputs = num_outputs

  def build(self, input_shape):
    self.kernel = self.add_weight("kernel",
                                  shape=[int(input_shape[-1]),
                                         self.num_outputs])

  def call(self, inputs):
    return tf.matmul(inputs, self.kernel)
    
  def get_config(self):
    return { "num_outputs": self.num_outputs }

layer = MyDenseLayer(10)

class ResnetIdentityBlock(tf.keras.Model):
  def get_config(self):
      return { "kernel_size": self.kernel_size, "filters": self.filters}
    
  def __init__(self, kernel_size, filters):
    super(ResnetIdentityBlock, self).__init__(name='ResnetIdentityBlock')
    self.kernel_size = kernel_size
    self.filters = filters
    filters1, filters2, filters3 = filters

    self.conv2a = tf.keras.layers.Conv2D(filters1, (1, 1))
    self.bn2a = tf.keras.layers.BatchNormalization()

    self.conv2b = tf.keras.layers.Conv2D(filters2, kernel_size, padding='same')
    self.bn2b = tf.keras.layers.BatchNormalization()

    self.conv2c = tf.keras.layers.Conv2D(filters3, (1, 1))
    self.bn2c = tf.keras.layers.BatchNormalization()

  def call(self, input_tensor, training=False):
    x = self.conv2a(input_tensor)
    x = self.bn2a(x, training=training)
    x = tf.nn.relu(x)

    x = self.conv2b(x)
    x = self.bn2b(x, training=training)
    x = tf.nn.relu(x)

    x = self.conv2c(x)
    x = self.bn2c(x, training=training)

    x += input_tensor
    return tf.nn.relu(x)

inp = Input((64, 64, 3))
block = ResnetIdentityBlock(1, [1, 2, 3])

model = tf.keras.Sequential([inp, Conv2D(1, 3), block, Conv2D(16, 5), layer])
dummy_input = np.zeros((1, 64, 64, 3))

autumn8.attach_model(model, dummy_input)



from typing import Optional

from pydantic import BaseModel

from autumn8.common.config import settings


class NewModelInfo(BaseModel):
    name: str
    s3_file_url: Optional[str] = None  # initialized later
    s3_input_file_url: Optional[str] = None
    framework: settings.Framework
    quantization: settings.Quantization
    model_file_type: Optional[str] = None
    domain: Optional[str] = None
    task: Optional[str] = None
    inputs: Optional[str] = None
    group_id: Optional[str] = None



import io
import json
import os
import sys
import zipfile

from autumn8.common._version import __version__
from autumn8.lib.package_resolver import lazy_import_pytorch, package

sys.path.append(os.path.join(os.path.dirname(__file__), "../"))


def export_tensorflow_model_repr(
    model,
    dummy_input,
    interns=[],
    externs=[],
    max_search_depth=5,
    preprocess=None,
    postprocess=None,
):
    return export_model_repr(
        model,
        dummy_input,
        "TENSORFLOW",
        interns,
        externs,
        max_search_depth,
        preprocess,
        postprocess,
    )


def export_pytorch_model_repr(
    model,
    dummy_input,
    interns=[],
    externs=[],
    max_search_depth=5,
    preprocess=None,
    postprocess=None,
):
    return export_model_repr(
        model,
        dummy_input,
        "PYTORCH",
        interns,
        externs,
        max_search_depth,
        preprocess,
        postprocess,
    )


def export_model_repr(
    model,
    dummy_input,
    framework,
    interns=[],
    externs=[],
    max_search_depth=5,
    preprocess=None,
    postprocess=None,
):
    bytes = io.BytesIO()

    if framework == "PYTORCH":
        externs.append("torch")
        externs.append("torchvision")

    if framework == "TENSORFLOW":
        externs.append("tensorflow")

    file, requirements, package_name = package(
        model, interns, externs, max_search_depth
    )

    (
        file_preprocess,
        requirements_preprocess,
        package_name_preprocess,
        file_postprocess,
        requirements_postprocess,
        package_name_postprocess,
    ) = [None for i in range(6)]

    if preprocess is not None:
        (
            file_preprocess,
            requirements_preprocess,
            package_name_preprocess,
        ) = package(preprocess, interns, externs, max_search_depth)
        requirements.update(requirements_preprocess)

    if postprocess is not None:
        (
            file_postprocess,
            requirements_postprocess,
            package_name_postprocess,
        ) = package(postprocess, interns, externs, max_search_depth)
        requirements.update(requirements_postprocess)

    (
        file_input,
        requirements_input,
        package_name_input,
    ) = package(dummy_input, interns, externs, max_search_depth)
    requirements.update(requirements_input)

    with zipfile.ZipFile(bytes, "w") as zip:
        zip.writestr(
            "MANIFEST",
            json.dumps(
                {
                    "version": __version__,
                    "package_name": package_name,
                    "package_name_preprocess": package_name_preprocess,
                    "package_name_postprocess": package_name_postprocess,
                    "package_name_input": package_name_input,
                }
            ),
        )
        zip.write(file.name, arcname="model.package")
        zip.write(file_input.name, arcname="input.package")
        if file_preprocess is not None:
            zip.write(file_preprocess.name, arcname="preprocess.package")
        if file_postprocess is not None:
            zip.write(file_postprocess.name, arcname="postprocess.package")
        requirement_list = []
        for package_name, package_version in requirements.items():
            requirement_list.append(f"{package_name}=={package_version}")

        zip.writestr("requirements.txt", "\n".join(requirement_list))

    bytes.seek(0)
    return bytes


def load_model(filename):
    extract_path = "/tmp/" + os.path.basename(filename)
    with zipfile.ZipFile(filename) as z:
        z.extractall(extract_path)

    with open(os.path.join(extract_path, "MANIFEST"), "r") as manifest:
        manifest_content = json.loads(manifest.read())
        package_name = manifest_content["package_name"]
        package_name_preprocess = (
            manifest_content["package_name_preprocess"]
            if "package_name_preprocess" in manifest_content
            else None
        )
        package_name_postprocess = (
            manifest_content["package_name_postprocess"]
            if "package_name_postprocess" in manifest_content
            else None
        )
        package_name_input = (
            manifest_content["package_name_input"]
            if "package_name_input" in manifest_content
            else None
        )
    torch = lazy_import_pytorch()
    model_file = open(os.path.join(extract_path, "model.package"), "rb")
    resource_name = "model.pkl"
    imp = torch.package.PackageImporter(model_file)
    loaded_model = imp.load_pickle(package_name, resource_name)

    loaded_input = None
    if package_name_input is not None:
        input_file = open(os.path.join(extract_path, "input.package"), "rb")
        resource_name = "model.pkl"
        imp = torch.package.PackageImporter(input_file)
        loaded_input = imp.load_pickle(package_name_input, resource_name)

    loaded_preprocess = None
    if package_name_preprocess is not None:
        model_file = open(
            os.path.join(extract_path, "preprocess.package"), "rb"
        )
        resource_name = "model.pkl"
        imp = torch.package.PackageImporter(model_file)
        loaded_preprocess = imp.load_pickle(
            package_name_preprocess, resource_name
        )

    loaded_postprocess = None
    if package_name_postprocess is not None:
        model_file = open(
            os.path.join(extract_path, "postprocess.package"), "rb"
        )
        resource_name = "model.pkl"
        imp = torch.package.PackageImporter(model_file)
        loaded_postprocess = imp.load_pickle(
            package_name_postprocess, resource_name
        )

    return loaded_model, loaded_preprocess, loaded_postprocess, loaded_input


attached_models = []


# TODO: create 'annotate' module?
def attach_model(
    model,
    example_input,
    interns=None,
    externs=None,
    max_search_depth=5,
    preprocess=None,
    postprocess=None,
):
    if interns is None:
        interns = []
    if externs is None:
        externs = []
    attached_models.append(
        (
            model,
            example_input,
            interns,
            externs,
            max_search_depth,
            preprocess,
            postprocess,
        )
    )



from typing import List, Literal, Optional, Union

import click
import questionary
from click import ClickException
from questionary import Choice

from autumn8.cli.cli_environment import CliEnvironment
from autumn8.cli.interactive import (
    pick_organization_id,
    verify_organization_id_access,
)
from autumn8.common.config.settings import CloudServiceProvider
from autumn8.common.config.supported_instances import find_instance_config
from autumn8.lib import logging

ENABLE_TOGGLEABLE_ENVIRONMENT = True


def use_environment(func):
    allowed_environments = (
        [env.name for env in CliEnvironment]
        if ENABLE_TOGGLEABLE_ENVIRONMENT
        else [CliEnvironment.PRODUCTION.name]
    )

    return click.option(
        "-e",
        "--environment",
        "--env",
        is_eager=True,  # often used when evaluating other options
        type=click.Choice(allowed_environments, case_sensitive=False),
        default=CliEnvironment.PRODUCTION.name,
        callback=lambda c, p, v: getattr(CliEnvironment, v),
        help="Environment to use",
        hidden=True,
    )(func)


def pick_or_verify_organization(ctx, param, value):
    organization_id = value
    environment = ctx.params["environment"]

    if organization_id is None:
        return pick_organization_id(environment)
    else:
        verify_organization_id_access(environment, organization_id)
        return value


use_organization_id = click.option(
    "-o",
    "--organization_id",
    "--org_id",
    type=int,
    callback=pick_or_verify_organization,
    help="The ID of the Organization to use",
)

use_auto_confirm = click.option(
    "-y",
    "--yes",
    "auto_confirm",
    type=bool,
    is_flag=True,
    help="Skip all confirmation input from the user.",
)

use_quiet_mode = click.option(
    "-q",
    "--quiet",
    is_flag=True,
    callback=lambda ctx, param, value: logging.set_max_log_verbosity(
        logging.ERROR
    )
    if value
    else None,
    help="Skip additional logging, printing only necessary info",
)


def pick_cloud_provider(options: List[CloudServiceProvider]):
    cloud_provider = questionary.select(
        "Choose service provider",
        choices=[
            Choice(title=provider.value, value=provider) for provider in options
        ],
        use_shortcuts=True,
    ).unsafe_ask()
    return cloud_provider


def get_valid_cloud_providers_for_machine_type(
    machine_type,
) -> List[CloudServiceProvider]:
    if machine_type is None:
        return list(CloudServiceProvider)

    instance_config = find_instance_config(
        machine_type, fetch_data_from_cloud_info=False
    )

    valid_providers = [
        instance_config.service_provider,
        CloudServiceProvider.AUTUMN8,
    ]

    return valid_providers


# TODO: update codebase to python 3.12+, unify these two
CloudProviderLabelExtensionsType = Literal[
    "a8f",
    "gcp",
    "aws",
    "None",
]
CloudProviderLabelExtensions: List[str] = [
    "a8f",
    "gcp",
    "aws",
    "None",
]

ExtendedCloudProviderLabel = Union[
    CloudProviderLabelExtensionsType,
    CloudServiceProvider,
]


def map_string_value_to_cloud_provider(
    value: str,
) -> Optional[CloudServiceProvider]:
    if value is None or value == str(None):
        return None

    if value.lower() == "gcp":
        return CloudServiceProvider.GOOGLE

    if value.lower() == "a8f":
        return CloudServiceProvider.AUTUMN8

    if value.lower() == "aws":
        return CloudServiceProvider.AMAZON

    for provider in CloudServiceProvider:
        if provider.value.lower() == value.lower():
            return provider

    raise ClickException(
        f"'{value}' is not a valid cloud provider. It must be one of "
        + f"{CloudProviderLabelExtensions+ [p.value for p in CloudServiceProvider] }",
    )


def get_callback_for_picking_valid_cloud_provider(
    optional: bool = False, default_value: Optional[CloudServiceProvider] = None
):
    def callback(ctx: click.Context, param: click.Parameter, value: str):
        provider_from_value = map_string_value_to_cloud_provider(value)

        machine_type = ctx.params.get("machine_type")
        valid_cloud_providers = get_valid_cloud_providers_for_machine_type(
            machine_type
        )

        if provider_from_value is None:
            if optional:
                return default_value
            return pick_cloud_provider(valid_cloud_providers)

        for provider in valid_cloud_providers:
            if provider is provider_from_value:
                return provider

        raise ClickException(
            f"{provider_from_value} is not a valid cloud provider for machine "
            + f"{machine_type}. It must be one of {valid_cloud_providers}",
        )

    return callback


def use_cloud_provider_picker(
    optional: bool = False, default_value: Optional[CloudServiceProvider] = None
):
    return click.option(
        "-c",
        "--cloud_provider",
        type=click.Choice(
            CloudProviderLabelExtensions
            + [p.value for p in CloudServiceProvider],
            case_sensitive=False,
        ),
        callback=get_callback_for_picking_valid_cloud_provider(
            optional=optional, default_value=default_value
        ),
        help="Cloud provider to use",
    )



from pathlib import Path

from setuptools import find_namespace_packages, setup

from autumn8.common._version import __version__

this_directory = Path(__file__).parent
readme_content = (this_directory / "README.md").read_text()

# all the configuration is now handled in pyproject.toml
setup()



# this is a wrapper for the `logging` module
# pyright: reportWildcardImportFromLibrary=false
# pylint: disable=wildcard-import,unused-wildcard-import

import logging
import logging.config
import os
from logging import *
from logging import Logger
from logging import getLogger as oldGetLogger
from typing import Optional

import yaml

with open(
    os.path.join(os.path.dirname(__file__), ".", "logging.yaml"),
    "r",
    encoding="utf-8",
) as f:
    config = yaml.safe_load(f.read())
    logging.config.dictConfig(config)


def getLogger(  # pylint: disable=invalid-name,function-redefined  # overriding a builtin function from 'logging'
    name: Optional[str] = None,
) -> Logger:
    logger = oldGetLogger(name)
    logger.setLevel(logging.DEBUG)
    return logger


def trigger_alarm_system(logger: Logger, alarm_reason: str = ""):
    logger.error("alarm: 500 Internal Server Error, reason=(%s)", alarm_reason)


def set_max_log_verbosity(level):
    basicConfig(level=level, force=True)

    loggers = [getLogger(name) for name in getLogger("root").manager.loggerDict]
    for logger in loggers:
        if logger.level < level:
            logger.setLevel(level)



import sys
import time

import click
import questionary
from questionary import Choice

from autumn8.cli import options, pending_uploads
from autumn8.cli.commands import cloud, models
from autumn8.cli.interactive import announce_model_upload_response, normalize_args
from autumn8.common._version import __version__
from autumn8.lib import api
from autumn8.lib.service import resume_upload_model

try:
    current_pending_uploads = pending_uploads.retrieve_pending_uploads()
except (ValueError, AttributeError):
    # data format stored is incompatible with the current version
    # hope that the pending upload will get cleaned up by the S3 lifecycle policy
    # TODO: add the S3 lifecycle policy for that, lol - https://docs.aws.amazon.com/AmazonS3/latest/userguide/mpu-abort-incomplete-mpu-lifecycle-config.html
    pending_uploads.forget_all_pending_uploads()
    current_pending_uploads = {}

# TODO: save costs by configuring https://docs.aws.amazon.com/AmazonS3/latest/userguide/mpu-abort-incomplete-mpu-lifecycle-config.html
# TODO: detect if the cached upload is still available on S3; forget it if it got cleaned up
for key in current_pending_uploads:
    resume_args = current_pending_uploads[key]
    click.echo(
        "You have pending upload of {}".format(
            resume_args["model_filepath_or_url"]
            if "model_filepath_or_url" in resume_args
            else resume_args[
                "model_file"
            ]  # backwards compat - TODO: remove this in May or later
        )
    )
    continue_upload = questionary.select(
        "Do you want to continue upload?",
        choices=[
            Choice(title="Yes", value="Y"),
            Choice(title="No", value="n"),
            Choice(title="Drop upload", value="drop"),
        ],
        use_shortcuts=True,
    ).unsafe_ask()

    if continue_upload == "" or continue_upload == "Y":
        model_upload_response = resume_upload_model(
            {**resume_args, **resume_args["kwargs"]}
        )
        announce_model_upload_response(model_upload_response)
        sys.exit(0)

    if continue_upload == "drop":
        pending_uploads.abort_and_forget_upload(resume_args["run_id"])


@options.use_environment
def test_connection(environment):
    """
    Test AutoDL connection with the current API key.
    Displays the user's email address upon successful connection.
    """
    user_data = api.user.fetch_user_data(environment)
    print(f"Hello! You're authenticated as {user_data['email']}")


@click.group(context_settings={"token_normalize_func": normalize_args})
@click.version_option(version=__version__)
def main():
    pass


main.command()(test_connection)

all_commands = []

for command in [
    *models.model_commands_group.commands.values(),
    *cloud.cloud_commands_group.commands.values(),
]:
    main.add_command(command)


if __name__ == "__main__":
    main()



"""
Parsed environment variables, reexported under the same names

Use this module instead of parsing manually:
- to avoid parse logic duplication,
- to make it much easier to find all references,
- to have IDE autofill support :D
"""

import os
